# This script is used by pyqbdipreload to generate a json file that can be parsed with CoverageSingleRun.from_json
# This needs to be fast which is why we cannot import tritondse and generate the CoverageSingleRun directly 
# (`import tritondse` adds ~0.3 s to the execution time of the script in my experience).

# built-in modules
import atexit
import bisect
import ctypes
import ctypes.util
import json
import lief
import os
import sys
import time

from collections import Counter
from dataclasses import dataclass
from typing import List, Optional, Tuple, Dict

# Third-party modules
import pyqbdi


setjmp_data = {}
libdl = None


class Dl_info(ctypes.Structure):
    _fields_ = [
        ('dli_fname', ctypes.c_char_p),
        ('dli_fbase', ctypes.c_void_p),
        ('dli_sname', ctypes.c_char_p),
        ('dli_saddr', ctypes.c_void_p)
    ]


def dladdr(addr):
    res = Dl_info()
    libdl.dladdr(ctypes.cast(addr, ctypes.c_void_p), ctypes.byref(res))

    return res.dli_sname


def is_symbol(name, addr):
    sname = dladdr(addr)
    sname = sname.decode() if sname else ""

    return sname == name


def hook_post_setjmp(vm, state, gpr, fpr, data):
    setjmp_data[data]['rip'] = gpr.rip

    if setjmp_data[data]['setjmp_cbk_id'] is not None:
        vm.deleteInstrumentation(setjmp_data[data]['setjmp_cbk_id'])
        setjmp_data[data]['setjmp_cbk_id'] = None

    return pyqbdi.CONTINUE


def hook_post_longjmp(vm, state, gpr, fpr, data):
    if data in setjmp_data:
        gpr.rip = setjmp_data[data]['rip']
    else:
        print(f'[FATAL] longjmp arg ({data:x}) not found!')
        sys.exit(-1)

    if setjmp_data[data]['longjmp_cbk_id'] is not None:
        vm.deleteInstrumentation(setjmp_data[data]['longjmp_cbk_id'])
        setjmp_data[data]['longjmp_cbk_id'] = None

    return pyqbdi.CONTINUE


def handle_exec_transfer_call(vm, state, gpr, fpr, data):
    if is_symbol("_setjmp", gpr.rip):
        arg1 = gpr.rdi      # get env argument.

        setjmp_data[arg1] = {}
        setjmp_data[arg1]['setjmp_cbk_id'] = vm.addVMEventCB(pyqbdi.EXEC_TRANSFER_RETURN, hook_post_setjmp, arg1)
    elif is_symbol("longjmp", gpr.rip):
        arg1 = gpr.rdi      # get env argument.

        setjmp_data[arg1]['longjmp_cbk_id'] = vm.addVMEventCB(pyqbdi.EXEC_TRANSFER_RETURN, hook_post_longjmp, arg1)

    return pyqbdi.CONTINUE


@dataclass
class CoverageTrace:
    strategy: str  # BLOCK, EDGE, etc..
    covered_instructions: Counter
    covered_items: List[Tuple[int, int, Optional[int]]]
    modules: Dict[str, int]
    trace: List[int]


@dataclass
class CoverageData:
    strategy: str  # BLOCK, EDGE, etc..
    branch_data: Optional[Tuple[int, int, int, bool, bool]]  # (temporary data): branch pc, true-branch, false-branch, is_taken, is_dynamic
    trace: CoverageTrace
    modules_base: List[int]
    pie: bool
    dump_trace: bool

    def to_relative(self, addr: int) -> int:
        if self.pie:
            return addr - self.modules_base[bisect.bisect_right(self.modules_base, addr)-1]
        else:
            return addr


def get_modules() -> Dict[str, int]:
    """ Retrieve modules base address to remove ASLR. """
    modules = {}
    for m in pyqbdi.getCurrentProcessMaps(True):
        if m.name in modules:
            modules[m.name] = min(m.range[0], modules[m.name])
        else:
            modules[m.name] = m.range[0]
        # print(f"{m.name}: {m.range}, {m.permission}")
    return modules


def get_module_bases() -> List[int]:
    """ Retrieve modules base address to remove ASLR. """
    return sorted(get_modules().values())


def write_coverage(covdata: CoverageData, output_file: str):
    """Write coverage into a file.
    """
    data = {
        "coverage_strategy": covdata.trace.strategy,
        "covered_instructions": covdata.trace.covered_instructions,
        "covered_items": covdata.trace.covered_items,
        "trace": covdata.trace.trace,
        "modules_base": covdata.trace.modules
    }
    with open(output_file, "w") as fd:
        json.dump(data, fd)


def register_instruction_coverage(vm, gpr, fpr, data: CoverageData):
    # inst_analysis = vm.getInstAnalysis(type=pyqbdi.AnalysisType.ANALYSIS_INSTRUCTION)

    # Save instruction covered
    rel_rip = data.to_relative(gpr.rip)
    data.trace.covered_instructions[rel_rip] += 1  # change to be portable

    if data.dump_trace:
        # Also save the trace
        data.trace.trace.append(rel_rip)

    return pyqbdi.CONTINUE


def register_basic_block_coverage(vm, evt, gpr, fpr, data: CoverageData):
    addr = evt.basicBlockStart

    # Process branch data in case there is one pending.
    if data.branch_data:
        # Unpack branch data.
        branch_addr, true_branch_addr, false_branch_addr, is_taken, is_dynamic = data.branch_data
        br_a, true_a, false_a = data.to_relative(branch_addr), data.to_relative(true_branch_addr), data.to_relative(false_branch_addr)

        # Check if the branch was taken.
        taken_a, not_taken_a = (true_a, false_a) if bool(addr == true_branch_addr) else (false_a, true_a)

        if is_dynamic:
            data.trace.covered_items.append((br_a, taken_a, None))
        else:
            data.trace.covered_items.append((br_a, taken_a, not_taken_a))

        # Clear branch data for next occurrence.
        data.branch_data = None
    else:
        pass
        # FIXME: There is a problem in that script is_dynamic can never be true !
        # FIXME: as it's never set to true. I feel like the else here is the case where its dynamic?

    return pyqbdi.CONTINUE


def register_branch_coverage(vm, gpr, fpr, data):
    inst_analysis = vm.getInstAnalysis(type=pyqbdi.AnalysisType.ANALYSIS_INSTRUCTION | pyqbdi.AnalysisType.ANALYSIS_OPERANDS)

    operand = inst_analysis.operands[0]

    branch_addr = inst_analysis.address
    false_branch_addr = inst_analysis.address + inst_analysis.instSize

    if operand.type == pyqbdi.OperandType.OPERAND_IMM:
        # FIXME: Isn't it assuming the jump is relative ?
        true_branch_addr = inst_analysis.address + inst_analysis.instSize + ctypes.c_longlong(operand.value).value
    else:
        raise Exception('Invalid operand type')

    # Save current branch data
    data.branch_data = (branch_addr, true_branch_addr, false_branch_addr, None, False)

    return pyqbdi.CONTINUE


def pyqbdipreload_on_run(vm, start, stop):
    global libdl

    s = time.time()

    # Load dl library.
    libdl_path = ctypes.util.find_library('dl')
    if libdl_path is None:
        raise Exception('Unable to found dl library')
    libdl = ctypes.cdll.LoadLibrary(libdl_path)
    libdl.dladdr.argtypes = (ctypes.c_void_p, ctypes.POINTER(Dl_info))

    # Read parameters.
    strat = os.getenv('PYQBDIPRELOAD_COVERAGE_STRATEGY', 'BLOCK')
    output = os.getenv('PYQBDIPRELOAD_OUTPUT_FILEPATH', 'a.cov')
    bool_trace = os.getenv('PYQBDIPRELOAD_DUMP_TRACE', 'False')
    bool_trace = True if bool_trace in ['true', 'True'] else False

    mods = get_modules()
    base_addresses = sorted(get_modules().values())
    covtrace = CoverageTrace(strat, Counter(), [], mods, [])

    # Open binary in LIEF to check if PIE or not
    p = lief.parse(sys.argv[0])

    coverage_data = CoverageData(strat, None, covtrace, base_addresses, p.is_pie, bool_trace)

    # Remove all instrumented modules except the main one.
    vm.removeAllInstrumentedRanges()
    vm.addInstrumentedModuleFromAddr(start)

    if coverage_data.strategy == 'BLOCK' or bool_trace:
        # Add callback on instruction execution.
        vm.addCodeCB(pyqbdi.PREINST, register_instruction_coverage, coverage_data)

    # Add callback on basic block entry.
    vm.addVMEventCB(pyqbdi.BASIC_BLOCK_ENTRY, register_basic_block_coverage, coverage_data)

    # Add callback on the JCC mnemonic.
    vm.addMnemonicCB('JCC', pyqbdi.InstPosition.POSTINST, register_branch_coverage, coverage_data)

    # Write coverage on exit.
    # TODO This does not work with bins that crash.
    atexit.register(write_coverage, coverage_data, output)

    # Add callback for handling calls to functions setjmp and longjmp.
    vm.addVMEventCB(pyqbdi.EXEC_TRANSFER_CALL, handle_exec_transfer_call, None)

    # Run program.
    print("Run start")
    vm.run(start, stop)
    print(f"Run finished: {time.time() - s:.02f}s")



from __future__ import annotations

# built-in imports
from pathlib import Path
from typing import Optional, Generator, Tuple
from collections import namedtuple

# third party
import lief

# local imports
from tritondse.types import PathLike, Addr, Architecture, Platform, ArchMode, Perm, Endian
from tritondse.loaders import Loader, LoadableSegment
import tritondse.logging

logger = tritondse.logging.get("loader")

_arch_mapper = {
    lief.ARCHITECTURES.ARM:   Architecture.ARM32,
    lief.ARCHITECTURES.ARM64: Architecture.AARCH64,
    lief.ARCHITECTURES.X86:   Architecture.X86
}

_plfm_mapper = {
    lief.EXE_FORMATS.ELF: Platform.LINUX,
    lief.EXE_FORMATS.PE: Platform.WINDOWS,
    lief.EXE_FORMATS.MACHO: Platform.MACOS
}


class Program(Loader):
    """
    Representation of a program (loaded in memory). This class is wrapping
    `LIEF <https://lief.quarkslab.com/doc/latest>`_ to represent a program
    and to provide all the features allowing to pseudo-load one regardless
    of its format.
    """

    EXTERN_SYM_BASE = 0x0f001000
    EXTERN_SYM_SIZE = 0x1000

    BASE_STACK = 0xf0000000
    END_STACK  = 0x70000000 # This is inclusive

    def __init__(self, path: PathLike):
        """
        :param path: Program path
        :type path: :py:obj:`tritondse.types.PathLike`
        :raise FileNotFoundError: if the file is not properly recognized by lief
                                  or in the wrong architecture
        """
        super(Program, self).__init__(path)
        self.path: Path = Path(path)  #: Binary file path
        if not self.path.is_file():
            raise FileNotFoundError(f"file {path} not found (or not a file)")

        self._binary = lief.parse(str(self.path))
        if self._binary is None:  # lief has not been able to parse it
            raise FileNotFoundError(f"file {path} not recognised by lief")

        self._arch = self._load_arch()
        if self._arch is None:
            raise FileNotFoundError(f"binary {path} architecture unsupported {self._binary.abstract.header.architecture}")

        try:
            self._plfm = _plfm_mapper[self._binary.format]
            # TODO: better refine for Android, iOS etc.
        except KeyError:
            self._plfm = None

        self._funs = {f.name: f for f in self._binary.concrete.functions}

    @property
    def name(self) -> str:
        """ Name of the loader"""
        return f"Program({self.path})"

    @property
    def endianness(self) -> Endian:
        return {lief.ENDIANNESS.LITTLE: Endian.LITTLE,
                lief.ENDIANNESS.BIG: Endian.BIG}[self._binary.abstract.header.endianness]

    @property
    def entry_point(self) -> Addr:
        """
        Program entrypoint address as defined in the binary headers

        :rtype: :py:obj:`tritondse.types.Addr`
        """
        return self._binary.entrypoint

    @property
    def architecture(self) -> Architecture:
        """
        Architecture enum representing program architecture.

        :rtype: Architecture
        """
        return self._arch

    @property
    def platform(self) -> Optional[Platform]:
        """
        Platform of the binary. Its solely based on the format
        of the file ELF, PE etc..

        :return: Platform
        """
        return self._plfm

    @property
    def format(self) -> lief.EXE_FORMATS:
        """
        Binary format. Supported formats by lief are: ELF, PE, MachO

        :rtype: lief.EXE_FORMATS
        """
        return self._binary.format

    def _load_arch(self) -> Optional[Architecture]:
        """
        Load architecture as an Architecture object.

        :return: Architecture or None if unsupported
        """
        arch = self._binary.abstract.header.architecture
        if arch in _arch_mapper:
            arch = _arch_mapper[arch]
            if arch == Architecture.X86:
                arch = Architecture.X86 if self._binary.abstract.header.is_32 else Architecture.X86_64
            return arch
        else:
            return None

    @property
    def relocation_enum(self):
        """
        LIEF relocation enum associated with the current
        architecture of the binary.

        :return: LIEF relocation enum
        :rtype: Union[lief.ELF.RELOCATION_AARCH64,
                      lief.ELF.RELOCATION_ARM,
                      lief.ELF.RELOCATION_PPC64,
                      lief.ELF.RELOCATION_PPC,
                      lief.ELF.RELOCATION_i386,
                      lief.ELF.RELOCATION_X86_64]
        """
        rel_map = {
            lief.ELF.ARCH.AARCH64: lief.ELF.RELOCATION_AARCH64,
            lief.ELF.ARCH.ARM:     lief.ELF.RELOCATION_ARM,
            lief.ELF.ARCH.PPC64:   lief.ELF.RELOCATION_PPC64,
            lief.ELF.ARCH.PPC:     lief.ELF.RELOCATION_PPC,
            lief.ELF.ARCH.i386:    lief.ELF.RELOCATION_i386,
            lief.ELF.ARCH.x86_64:  lief.ELF.RELOCATION_X86_64
        }
        return rel_map[self._binary.concrete.header.machine_type]

    def _is_glob_dat(self, rel: lief.ELF.Relocation) -> bool:
        """ Get whether the given relocation is of type GLOB_DAT.
        Used locally to find mandatory relocations
        """
        rel_enum = self.relocation_enum
        if hasattr(rel_enum, "GLOB_DAT"):
            return rel_enum(rel.type) == getattr(rel_enum, "GLOB_DAT")
        else:
            return False  # Not GLOB_DAT relocation for this architecture

    def memory_segments(self) -> Generator[LoadableSegment, None, None]:
        """
        Iterate over all memory segments of the program as loaded in memory.

        :return: Generator of tuples addrs and content
        :raise NotImplementedError: if the binary format cannot be loaded
        """
        if self.format == lief.EXE_FORMATS.ELF:
            for i, seg in enumerate(self._binary.concrete.segments):
                if seg.type == lief.ELF.SEGMENT_TYPES.LOAD:
                    content = bytearray(seg.content)
                    if seg.virtual_size != len(seg.content):  # pad with zeros (as it might be .bss)
                        content += bytearray([0]) * (seg.virtual_size - seg.physical_size)
                    yield LoadableSegment(seg.virtual_address, perms=Perm(int(seg.flags)), content=bytes(content), name=f"seg{i}")
        else:
            raise NotImplementedError(f"memory segments not implemented for: {self.format.name}")

        # Also return a specific map to put external symbols
        yield LoadableSegment(self.EXTERN_SYM_BASE, self.EXTERN_SYM_SIZE, Perm.R | Perm.W, name="[extern]")
        yield LoadableSegment(self.END_STACK, self.BASE_STACK-self.END_STACK, Perm.R | Perm.W, name="[stack]")

    def imported_functions_relocations(self) -> Generator[Tuple[str, Addr], None, None]:
        """
        Iterate over all imported functions by the program. This function
        is a generator of tuples associating the function and its relocation
        address in the binary.

        :return: Generator of tuples function name and relocation address
        """
        if self.format == lief.EXE_FORMATS.ELF:
            try:
                # Iterate functions imported through PLT
                for rel in self._binary.concrete.pltgot_relocations:
                    yield rel.symbol.name, rel.address

                # Iterate functions imported via mandatory relocation (e.g: __libc_start_main)
                for rel in self._binary.dynamic_relocations:
                    if self._is_glob_dat(rel) and rel.has_symbol and not rel.symbol.is_variable:
                        yield rel.symbol.name, rel.address
            except Exception:
                logger.error('Something wrong with the pltgot relocations')

        else:
            raise NotImplementedError(f"Imported functions relocations not implemented for: {self.format.name}")

    def imported_variable_symbols_relocations(self) -> Generator[Tuple[str, Addr], None, None]:
        """
        Iterate over all imported variable symbols. Yield for each of them the name and
        the relocation address in the binary.

        :return: Generator of tuples with symbol name, relocation address
        """
        if self.format == lief.EXE_FORMATS.ELF:
            rel_enum = self.relocation_enum
            # Iterate imported symbols
            for rel in self._binary.dynamic_relocations:
                if rel.has_symbol:
                #if rel_enum(rel.type) == rel_enum.COPY and rel.has_symbol:
                    if rel.symbol.is_variable:
                        yield rel.symbol.name, rel.address
        else:
            raise NotImplementedError(f"Imported symbols relocations not implemented for: {self.format.name}")

    def find_function_addr(self, name: str) -> Optional[Addr]:
        """
        Search for the function name in fonctions of the binary.

        :param name: Function name
        :type name: str
        :return: Address of function if found
        :rtype: Addr
        """
        f = self._funs.get(name)
        return f.address if f else None

    @property
    def arch_mode(self) -> ArchMode:
        pass  # TODO: Richard



class SkipInstructionException(Exception):
    """
    Exception to raise in a PRE callback to skip the evaluation
    of the current instruction. It will thus force a SymbolicExecutor
    to fetch the next instruction. Thus the user have to update the
    RIP of the ProcessState currently being executed.
    """
    pass


class AbortExecutionException(Exception):
    """
    Exception to rais in a callback to stop current SymbolicExecutor.
    The user should be careful to set the status of the current seed
    being executed.
    """
    pass


class StopExplorationException(Exception):
    """
    Exception to raise in a callback to stop the whole exploration of
    the program. It is caught by SymbolicExplorator.
    """
    pass


class AllocatorException(Exception):
    """
    Class used to represent an heap allocator exception.
    This exception can be raised in the following conditions:

    * trying to allocate data which overflow heap size
    * trying to free a pointer already freed
    * trying to free a non-allocated pointer
    """
    def __init__(self, message):
        super(Exception, self).__init__(message)


class ProbeException(Exception):
    """
    Exception to raise in a probe to stop the current exception.
    It is caught by SymbolicExplorator.
    """
    def __init__(self, message):
        super(Exception, self).__init__(message)



from .config                import Config
from .loaders.program       import Program
from .loaders.quokkaprogram import QuokkaProgram
from .loaders.cle_loader    import CleLoader
from .loaders.loader        import Loader, MonolithicLoader, LoadableSegment
from .process_state         import ProcessState
from .coverage              import CoverageStrategy, BranchSolvingStrategy, CoverageSingleRun, GlobalCoverage
from .symbolic_executor     import SymbolicExecutor
from .symbolic_explorator   import SymbolicExplorator, ExplorationStatus
from .seed                  import Seed, SeedStatus, SeedFormat, CompositeData
from .sanitizers            import CbType, ProbeInterface, UAFSanitizer, NullDerefSanitizer, FormatStringSanitizer, IntegerOverflowSanitizer
from .types                 import *
from .workspace             import Workspace
from .memory                import Memory, MemoryAccessViolation, MapOverlapException, MemMap
from .exception import AllocatorException, SkipInstructionException, StopExplorationException, AbortExecutionException

from triton import VERSION

TRITON_VERSION = f"v{VERSION.MAJOR}.{VERSION.MINOR}"



from tritondse.types import Addr, ByteSize, Perm
from tritondse.memory import Memory
from tritondse.exception import AllocatorException
import tritondse.logging

logger = tritondse.logging.get()


class HeapAllocator(object):
    """
    Custom tiny heap allocator. Used by built-ins routines like malloc/free.
    This allocation manager also provides an API enabling checking whether
    a pointer is allocated freed etc.

    .. warning:: This allocator is very simple and does not perform any
                 coalescing of freed memory areas. Thus it may not correctly
                 model the behavior of libc allocator.
    """

    def __init__(self, start: Addr, end: Addr, memory: Memory):
        """
        Class constructor. Takes heap bounds as parameter.

        :param start Addr: Where the heap area can start
        :type start: :py:obj:`tritondse.types.Addr`
        :param end Addr: Where the heap area must be end
        :type start: :py:obj:`tritondse.types.Addr`
        :param memory: Memory: Memory object on which to perform allocations
        """
        # Range of the memory mapping
        self.start: Addr = start
        #: Starting address of the heap

        self.end: Addr = end
        #: Ending address of the heap

        self._curr_offset: Addr = self.start  #: Heap current offset address
        self._memory = memory

        # Pools memory
        self.alloc_pool = dict() # {ptr: MemMap}
        self.free_pool = dict() # {size: set(MemMap ...)}

        # TODO: For a to-the-moon allocator, we could merge freed chunks. Like 4 chunks of 1 byte into one chunk of 4 bytes.
        # TODO: For a to-the-moon allocator, we could split a big chunk into two chunks when asking an allocation.

    def alloc(self, size: ByteSize) -> Addr:
        """
        Performs an allocation of the given byte size.

        :param size: Byte size to allocate
        :type size: :py:obj:`tritondse.types.ByteSize`
        :raise AllocatorException: if not memory is available
        :return: The pointer address allocated
        :rtype: :py:obj:`tritondse.types.Addr`
        """

        if size <= 0:
            logger.error(f"Heap: invalid allocation size {size}")
            return 0

        ptr = None
        for sz in sorted(x for x in self.free_pool if x >= size):
            # get the free chunk
            ptr = self.free_pool[sz].pop().start

            # If the set is empty after the pop(), remove the entry
            if not self.free_pool[sz]:
                del self.free_pool[sz]
            break

        if ptr is None: # We did not found reusable freed ptr
            ptr = self._curr_offset
            self._curr_offset += size

        # Now we can allocate the chunk
        map = self._memory.map(ptr, size, Perm.R | Perm.W, 'heap')
        self.alloc_pool.update({ptr: map})

        return ptr

    def free(self, ptr: Addr) -> None:
        """
        Free the given memory chunk.

        :param ptr: Address to free
        :type ptr: :py:obj:`tritondse.types.Addr`
        :raise AllocatorException: if the pointer has already been freed or if it has never been allocated
        """
        if self.is_ptr_freed(ptr):
            raise AllocatorException('Double free or corruption!')

        if not self.is_ptr_allocated(ptr):
            raise AllocatorException(f'Invalid pointer ({hex(ptr)})')

        # Add the chunk into our free_pool
        memmap = self.alloc_pool[ptr]
        if memmap.size in self.free_pool:
            self.free_pool[memmap.size].add(memmap)
        else:
            self.free_pool[memmap.size] = {memmap}

        # Remove the chunk from our alloc_pool
        self._memory.unmap(ptr)
        del self.alloc_pool[ptr]

    def is_ptr_allocated(self, ptr: Addr) -> bool:
        """
        Check whether a given address has been allocated

        :param ptr: Address to check
        :type ptr: :py:obj:`tritondse.types.Addr`
        :return: True if pointer points to an allocated memory region
        """
        return self._memory.is_mapped(ptr, 1)

    def is_ptr_freed(self, ptr: Addr) -> bool:
        """
        Check whether a given pointer has recently been freed.

        :param ptr: Address to check
        :type ptr: :py:obj:`tritondse.types.Addr`
        :return: True if pointer has been freed, False otherwise
        """
        # FIXME: This function is linear in the size of chunks. Can make it logarithmic
        for size, chunks in self.free_pool.items():
            for chunk in chunks:
                if chunk.start <= ptr < chunk.start + size:
                    return True
        return False



from __future__ import annotations

# built-in imports
from pathlib import Path
from typing import Optional, Generator, Tuple, Dict, List
from dataclasses import dataclass

# local imports
from tritondse.types import Addr, Architecture, Platform, ArchMode, Perm, Endian
from tritondse.arch import ARCHS
import tritondse.logging

logger = tritondse.logging.get()


@dataclass
class LoadableSegment:
    """ Represent a Segment to load in memory.
    It can either provide a content and will thus be
    initialized in a context or virtual.
    """
    address: int
    """ Virtual address where to load the segment """
    size: int = 0
    """ Size of the segment. If content is present use len(content)"""
    perms: Perm = Perm.R|Perm.W|Perm.X
    """ Permissions to assign the segment """
    content: Optional[bytes] = None
    """ Content of the segment """
    name: str = ""
    """ Name to give to the segment """


class Loader(object):
    """
    This class describes how to load the target program in memory.
    """
    def __init__(self, path: str):
        self.bin_path = Path(path)

    @property
    def name(self) -> str:
        """
        Name of the loader and target being loaded.

        :return: str of the loader name
        """
        raise NotImplementedError()

    @property
    def entry_point(self) -> Addr:
        """
        Program entrypoint address as defined in the binary headers

        :rtype: :py:obj:`tritondse.types.Addr`
        """
        raise NotImplementedError()

    @property
    def architecture(self) -> Architecture:
        """
        Architecture enum representing program architecture.

        :rtype: Architecture
        """
        raise NotImplementedError()

    @property
    def arch_mode(self) -> Optional[ArchMode]:
        """
        ArchMode enum representing the starting mode (e.g Thumb for ARM).
        if None, the default mode of the architecture will be used.

        :rtype: Optional[ArchMode]
        """
        return None

    @property
    def platform(self) -> Optional[Platform]:
        """
        Platform of the binary.

        :return: Platform
        """
        return None

    @property
    def endianness(self) -> Endian:
        """
        Endianess of the loaded program

        :return: Endianness
        """
        raise NotImplementedError()

    def memory_segments(self) -> Generator[LoadableSegment, None, None]:
        """
        Iterate over all memory segments of the program as loaded in memory.

        :return: Generator of tuples addrs and content
        :raise NotImplementedError: if the binary format cannot be loaded
        """
        raise NotImplementedError()

    @property
    def cpustate(self) -> Dict[str, int]:
        """
        Provide the initial cpu state in the forma of a dictionary of
        {"register_name" : register_value}
        """
        return {}

    def imported_functions_relocations(self) -> Generator[Tuple[str, Addr], None, None]:
        """
        Iterate over all imported functions by the program. This function
        is a generator of tuples associating the function and its relocation
        address in the binary.

        :return: Generator of tuples function name and relocation address
        """
        yield from ()

    def imported_variable_symbols_relocations(self) -> Generator[Tuple[str, Addr], None, None]:
        """
        Iterate over all imported variable symbols. Yield for each of them the name and
        the relocation address in the binary.

        :return: Generator of tuples with symbol name, relocation address
        """
        yield from ()

    def find_function_addr(self, name: str) -> Optional[Addr]:
        """
        Search for the function name in fonctions of the binary.

        :param name: Function name
        :type name: str
        :return: Address of function if found
        :rtype: Addr
        """
        return None


class MonolithicLoader(Loader):
    """
    Monolithic loader. It helps loading raw firmware at a given address
    in DSE memory space, with the various attributes like architecture etc.
    """

    def __init__(self,
                 architecture: Architecture,
                 cpustate: Dict[str, int] = None,
                 maps: List[LoadableSegment] = None,
                 set_thumb: bool = False,
                 platform: Platform = None,
                 endianess: Endian = Endian.LITTLE):
        super(MonolithicLoader, self).__init__("")

        self._architecture = architecture
        self._platform = platform if platform else None
        self._cpustate = cpustate if cpustate else {}
        self.maps = maps
        self._arch_mode = ArchMode.THUMB if set_thumb else None
        self._endian = endianess
        if self._platform and (self._architecture, self._platform) in ARCHS:
            self._archinfo = ARCHS[(self._architecture, self._platform)]
        elif self._architecture in ARCHS:
            self._archinfo = ARCHS[self._architecture]
        else: 
            logger.error("Unknown architecture")
            assert False

    @property
    def name(self) -> str:
        """ Name of the loader"""
        return f"Monolithic({self.bin_path})"

    @property
    def architecture(self) -> Architecture:
        """
        Architecture enum representing program architecture.

        :rtype: Architecture
        """
        return self._architecture

    @property
    def arch_mode(self) -> ArchMode:
        """
        ArchMode enum representing the starting mode (e.g Thumb for ARM).

        :rtype: ArchMode
        """
        return self._arch_mode

    @property
    def entry_point(self) -> Addr:
        """
        Program entrypoint address as defined in the binary headers

        :rtype: :py:obj:`tritondse.types.Addr`
        """
        return self.cpustate[self._archinfo.pc_reg]

    def memory_segments(self) -> Generator[LoadableSegment, None, None]:
        """
        In the case of a monolithic firmware, there is a single segment.
        The generator returns a single tuple with the load address and the content.

        :return: Generator of tuples addrs and content
        """
        yield from self.maps

    @property
    def cpustate(self) -> Dict[str, int]:
        """
        Provide the initial cpu state in the format of a dictionary of
        {"register_name" : register_value}
        """
        return self._cpustate

    @property
    def platform(self) -> Optional[Platform]:
        """
        Platform of the binary.

        :return: Platform
        """
        return self._platform

    @property
    def endianness(self) -> Endian:
        """
        Endianess of the monolithic loader.
        (default is LITTLE)
        """
        return self._endian



# built-in imports
import io
import time
import os

if os.name == 'posix':
    import resource

from typing import Optional, Union, List, NoReturn, Dict, Type

# third party imports
from triton import MODE, Instruction, CPUSIZE, MemoryAccess, CALLBACK

# local imports
from tritondse.config import Config
from tritondse.coverage import CoverageSingleRun, BranchSolvingStrategy
from tritondse.process_state import ProcessState
from tritondse.loaders import Loader
from tritondse.seed import Seed, SeedStatus, SeedFormat, CompositeData
from tritondse.types import Expression, Architecture, Addr, Model, SymbolicVariable, Register
from tritondse.routines import SUPPORTED_ROUTINES, SUPORTED_GVARIABLES
from tritondse.callbacks import CallbackManager
from tritondse.workspace import Workspace
from tritondse.heap_allocator import AllocatorException
from tritondse.thread_context import ThreadContext
from tritondse.exception import AbortExecutionException, SkipInstructionException, StopExplorationException, ProbeException
from tritondse.memory import MemoryAccessViolation, Perm
import tritondse.logging

logger = tritondse.logging.get("executor")


class SymbolicExecutor(object):
    """
    Single Program Execution Class.
    That module, is in charge of performing the process loading from the given
    program.
    """

    def __init__(self, config: Config, seed: Seed = Seed(), workspace: Workspace = None, uid=0, callbacks=None):
        """
        :param config: configuration file to use
        :type config: Config
        :param seed: input file to inject either in stdin or argv (optional)
        :type seed: Seed
        :param workspace: Workspace to use. If None it will be instanciated
        :type workspace: Optional[Workspace]
        :param uid: Unique ID. Given by :py:obj:`SymbolicExplorator` to identify uniquely executions
        :type uid: int
        :param callbacks: callbacks to bind on this execution before running *(instanciated if empty !)*
        :type callbacks: CallbackManager
        """
        self.config: Config = config      #: Configuration file used
        self.loader: Type[Loader] = None  #: Loader used to run the code

        self.pstate: ProcessState = None  #: ProcessState

        self.workspace: Workspace = workspace  #: Current workspace
        if self.workspace is None:
            self.workspace = Workspace(config.workspace)

        self.seed: Seed = seed  #: The current seed used for the execution

        # Override config if there is a mismatch between seed format and config file
        if seed.format != self.config.seed_format:
            logger.warning(f"seed format {seed.format} mismatch config {config.seed_format} (override config)")
            self.config.seed_format = seed.format

        self.symbolic_seed = self._init_symbolic_seed(seed) #: symbolic seed (same structure than Seed but with symbols)

        self.coverage: CoverageSingleRun = CoverageSingleRun(self.config.coverage_strategy)  #: Coverage of the execution
        self.rtn_table = dict()   # Addr -> Tuple[fname, routine]
        self.uid: int = uid       #: Unique identifier meant to unique accross Exploration instances
        self.start_time: int = 0  #: start time of the process
        self.end_time: int = 0    #: end time of the process

        # create callback object if not provided as argument, and bind callbacks to the current process state
        self.cbm: CallbackManager = callbacks if callbacks is not None else CallbackManager()
        """callback manager"""

        # List of new seeds filled during the execution and flushed by explorator
        self._pending_seeds = []
        self._run_to_target = None

        self.trace_offset: int = 0  #: counter of instructions executed

        self.previous_pc: int = 0  #: previous program counter executed
        self.current_pc = 0        #: current program counter

        self.debug_pp = False

        self._in_processing = False  # use to know if we are currently processing an instruction

        # TODO: Here we load the binary each time we run an execution (via ELFLoader). We can
        #       avoid this (and so gain in speed) if a TritonContext could be forked from a
        #       state. See: https://github.com/JonathanSalwan/Triton/issues/532

    def _init_symbolic_seed(self, seed: Seed) -> Union[list, CompositeData]:
        if seed.is_raw():
            return [None]*len(seed.content)
        else:  # is composite
            argv = [[None]*len(a) for a in seed.content.argv]
            files = {k: [None]*len(v) for k, v in seed.content.files.items()}
            variables = {k: [None]*(1 if isinstance(v, int) else len(v)) for k, v in seed.content.variables.items()}
            return CompositeData(argv=argv, files=files, variables=variables)

    def load(self, loader: Loader) -> None:
        """
        Use the given loader to initialize the ProcessState.
        It overrides the current ProcessState if any.

        :param loader: Loader describing how to load
        :return: None
        """

        # Initialize the process_state architecture (at this point arch is sure to be supported)
        self.loader = loader
        logger.debug(f"Loading program {self.loader.name} [{self.loader.architecture}]")
        self.pstate = ProcessState.from_loader(loader)
        self._map_dynamic_symbols()
        self._load_seed_process_state(self.pstate, self.seed)

    def load_process(self, pstate: ProcessState) -> None:
        """
        Load the given process state. Do nothing but
        setting the internal ProcessState.

        :param pstate: PrcoessState to set
        """
        self.pstate = pstate
        self._load_seed_process_state(self.pstate, self.seed)

    @staticmethod
    def _load_seed_process_state(pstate: ProcessState, seed: Seed) -> None:
        if seed.is_raw():
            data = seed.content
        else:  # is composite
            if seed.is_file_defined("stdin"):
                data = seed.get_file_input("stdin")
            else:
                return
        filedesc = pstate.get_file_descriptor(0)
        filedesc.fd = io.BytesIO(data)

    @property
    def execution_time(self) -> int:
        """
        Time taken for the execution in seconds

        .. warning:: Only relevant at the end of the execution

        :return: execution time (in s)
        """
        return self.end_time - self.start_time

    @property
    def pending_seeds(self) -> List[Seed]:
        """
        List of pending seeds gathered during execution.

        .. warning:: Only relevant at the end of execution

        :returns: list of new seeds generated
        :rtype: List[Seed]
        """
        return self._pending_seeds

    def enqueue_seed(self, seed: Seed) -> None:
        """
        Add a seed to the queue of seed to be executed in later iterations.
        This function is meant to be used by user callbacks.

        :param seed: Seed to be added
        :type seed: Seed
        """
        self._pending_seeds.append(seed)

    @property
    def callback_manager(self) -> CallbackManager:
        """
        Get the callback manager associated with the execution.

        :rtype: CallbackManager"""
        return self.cbm

    def is_seed_injected(self) -> bool:
        """
        Get whether or not the seed has been injected.

        :return: True if the seed has already been inserted
        """
        if self.config.is_format_raw():
            return bool(self.symbolic_seed)
        elif self.config.is_format_composite():
            # Namely has one of the various input been injected or not
            return bool(self.symbolic_seed.content.files) or bool(self.symbolic_seed.content.variables)
        else:
            assert False

    def _configure_pstate(self) -> None:
        #for mode in [MODE.ALIGNED_MEMORY, MODE.AST_OPTIMIZATIONS, MODE.CONSTANT_FOLDING, MODE.ONLY_ON_SYMBOLIZED]:
        for mode in [MODE.ONLY_ON_SYMBOLIZED]:
            self.pstate.set_triton_mode(mode, True)
        logger.info(f"configure pstate: time_inc:{self.config.time_inc_coefficient}  solver:{self.config.smt_solver.name}  timeout:{self.config.smt_timeout}")
        self.pstate.time_inc_coefficient = self.config.time_inc_coefficient
        self.pstate.set_solver_timeout(self.config.smt_timeout)
        self.pstate.set_solver(self.config.smt_solver)

    def _fetch_next_thread(self, threads: List[ThreadContext]) -> Optional[ThreadContext]:
        """
        Given a list of threads, returns the next to execute. Iterating
        threads in a round-robin style picking the next item in the list.

        :param threads: list of threads
        :return: thread context
        """
        cur_idx = threads.index(self.pstate.current_thread)

        tmp_list = threads[cur_idx+1:]+threads[:cur_idx]  # rotate list (and exclude current_item)
        for th in tmp_list:
            if th.is_running():
                return th  # Return the first thread that is properly running
        return None

    def __schedule_thread(self) -> None:
        threads_list = self.pstate.threads

        if len(threads_list) == 1:  # If there is only one thread no need to schedule another thread
            return

        if self.pstate.current_thread.count > self.config.thread_scheduling:
            # Select the next thread to execute
            next_th = self._fetch_next_thread(threads_list)

            if next_th:  # We found another thread to schedule

                # Call all callbacks related to threads
                for cb in self.cbm.get_context_switch_callback():
                    cb(self, self.pstate, self.pstate.current_thread)

                # Save current context and restore new thread context (+kill current if dead)
                self.pstate.switch_thread(next_th)

            else:  # There are other thread but not other one is available (thus keep current one)
                self.pstate.current_thread.count = 0  # Reset its counter

        else:
            # Increment the instruction counter of the thread (a bit in advance but it does not matter)
            self.pstate.current_thread.count += 1

    def _symbolic_mem_callback(self, se: 'SymbolicExecutor', ps: ProcessState, mem: MemoryAccess, *args):
        tgt_addr = mem.getAddress()
        lea_ast = mem.getLeaAst()
        if lea_ast is None:
            return
        if lea_ast.isSymbolized():
            s = "write" if bool(args) else "read"
            pc = self.pstate.cpu.program_counter
            logger.debug(f"symbolic {s} at 0x{pc:x}: target: 0x{tgt_addr:x} [{lea_ast}]")
            self.pstate.push_constraint(lea_ast == tgt_addr, f"sym-{s}:{self.trace_offset}:{pc}")

    def emulate(self):
        while not self.pstate.stop and self.pstate.threads:
            if not self.step():
                break
        if not self.seed.is_status_set():  # Set a status if it has not already been done
            self.seed.status = SeedStatus.OK_DONE
        return


    def step(self) -> bool:
        """
        Perform a single instruction step. Returns whether the emulation can
        continue or we have to stop.
        """
        try:
            # Schedule thread if it's time
            self.__schedule_thread()

            if not self.pstate.current_thread.is_running():
                logger.warning(f"After scheduling current thread is not running (probably in a deadlock state)")
                return False  # Were not able to find a suitable thread thus exit emulation

            # Fetch program counter (of the thread selected), at this point the current thread should be running!
            self.current_pc = self.pstate.cpu.program_counter  # should normally be already set but still.

            if self.current_pc == self._run_to_target:  # Hit the location we wanted to reach
                return False

            if self.current_pc == 0:
                logger.error(f"PC=0, is it normal ? (stop)")
                return False

            if self.pstate.memory.segmentation_enabled:
                if not self.pstate.memory.has_ever_been_written(self.current_pc, CPUSIZE.BYTE):
                    logger.error(f"Instruction not mapped: 0x{self.current_pc:x}")
                    return False

            instruction = self.pstate.fetch_instruction()
            opcode = instruction.getOpcode()
            mnemonic = instruction.getType()

            try:
                # Trigger pre-address callback
                pre_cbs, post_cbs = self.cbm.get_address_callbacks(self.current_pc)
                for cb in pre_cbs:
                    cb(self, self.pstate, self.current_pc)

                # Trigger pre-opcode callback
                pre_opcode, post_opcode = self.cbm.get_opcode_callbacks(opcode)
                for cb in pre_opcode:
                    cb(self, self.pstate, opcode)

                # Trigger pre-mnemonic callback
                pre_mnemonic, post_mnemonic = self.cbm.get_mnemonic_callbacks(mnemonic)
                for cb in pre_mnemonic:
                    cb(self, self.pstate, mnemonic)

                # Trigger pre-instruction callback
                pre_insts, post_insts = self.cbm.get_instruction_callbacks()
                for cb in pre_insts:
                    cb(self, self.pstate, instruction)
            except SkipInstructionException as _:
                return True

            if self.pstate.is_syscall():
                logger.warning(f"execute syscall instruction {self.pstate.read_register(self.pstate._syscall_register)}")

            # Process
            prev_pc = self.current_pc
            self._in_processing = True
            if not self.pstate.process_instruction(instruction):
                if self.pstate.is_halt_instruction():
                    logger.info(f"hit {str(instruction)} instruction stop.")
                    return False
                else:
                    logger.error('Instruction not supported: %s' % (str(instruction)))
                    if self.config.skip_unsupported_instruction:
                        self.pstate.cpu.program_counter += instruction.getSize() # try to jump over the instruction
                    else:
                        return False  # stop emulation

            self._in_processing = False
            # increment trace offset
            self.trace_offset += 1

            # update previous program counters
            self.previous_pc = prev_pc
            self.current_pc = self.pstate.cpu.program_counter  # current_pc becomes new instruction pointer

            # Update the coverage of the execution
            self.coverage.add_covered_address(self.previous_pc)

            # Update coverage send it the last PathConstraint object if one was added
            if self.pstate.is_path_predicate_updated():
                path_constraint = self.pstate.last_branch_constraint

                if path_constraint.isMultipleBranches():
                    branches = path_constraint.getBranchConstraints()
                    if len(branches) != 2:
                        logger.error("Branching condition has more than two branches")
                    taken, not_taken = branches if branches[0]['isTaken'] else branches[::-1]
                    taken_addr, not_taken_addr = taken['dstAddr'], not_taken['dstAddr']

                    for cb in self.cbm.get_on_branch_covered_callback():
                        cb(self, self.pstate, (self.previous_pc, taken_addr))

                    self.coverage.add_covered_branch(self.previous_pc, taken_addr, not_taken_addr)

                else:  # It is normally a dynamic jump or symbolic memory read/write
                    cmt = path_constraint.getComment()
                    if cmt.startswith("sym-read") or cmt.startswith("sym-write"):
                        pass
                        # NOTE: At the moment it does not seems suitable to count r/w pointers
                        # as part of the coverage. So does not have an influence on covered/not_covered.
                    else:
                        logger.warning(f"New dynamic jump covered at: {self.previous_pc:08x}")
                        path_constraint.setComment(f"dyn-jmp:{self.trace_offset}:{self.previous_pc}")
                        self.coverage.add_covered_dynamic_branch(self.previous_pc, self.current_pc)

            # Trigger post-opcode callback
            for cb in post_opcode:
                cb(self, self.pstate, opcode)

            # Trigger post-mnemonic callback
            for cb in post_mnemonic:
                cb(self, self.pstate, mnemonic)

            # Trigger post-instruction callback
            for cb in post_insts:
                cb(self, self.pstate, instruction)

            # Trigger post-address callbacks
            for cb in post_cbs:
                cb(self, self.pstate, self.previous_pc)

            # Simulate routines
            try:
                self._routines_handler(instruction)
            except AllocatorException as e:
                logger.info(f'An exception has been raised: {e}')
                self.seed.status = SeedStatus.CRASH
                return False

            # Check timeout of the execution
            if self.config.execution_timeout and (time.time() - self.start_time) >= self.config.execution_timeout:
                logger.info('Timeout of an execution reached')
                self.seed.status = SeedStatus.HANG
                return False
            return True

            # Call all the callbacks on the memory violations
            for cb in self.callback_manager.get_memory_violation_callbacks():
                cb(self, self.pstate, e)
        except AbortExecutionException as e:
            return False
        except MemoryAccessViolation as e:
            logger.warning(f"Memory violation: {str(e)}")
        except ProbeException:
            return False
        except Exception as e:
            logger.warning(f"Execution interrupted: {e}")
            self.seed.status = SeedStatus.FAIL
            return False

            # Assign the seed the status of crash
            if not self.seed.is_status_set():
                self.seed.status = SeedStatus.CRASH
            return False


    def __handle_external_return(self, routine_name: str, ret_val: Optional[Union[int, Expression]]) -> None:
        """ Symbolize or concretize return values of external functions """
        if ret_val is not None:
            reg = self.pstate.return_register
            if isinstance(ret_val, int): # Write its concrete value
                self.pstate.write_register(reg, ret_val)
            else:  # It should be a logic expression
                self.pstate.write_symbolic_register(reg, ret_val, f"(routine {routine_name}")

    def _routines_handler(self, instruction: Instruction):
        """
        This function handle external routines calls. When the .plt jmp on an external
        address, we call the appropriate Python routine and setup the returned value
        which may be concrete or symbolic.

        :param instruction: The current instruction executed
        :return: None
        """
        pc = self.pstate.cpu.program_counter
        if pc in self.rtn_table:
            routine_name, routine = self.rtn_table[pc]
            logger.debug(f"Enter external routine: {routine_name}")

            # Trigger pre-address callback
            pre_cbs, post_cbs = self.cbm.get_imported_routine_callbacks(routine_name)

            ret_val = None
            for cb in pre_cbs:
                ret = cb(self, self.pstate, routine_name, pc)
                if ret is not None:  # if the callback return a value the function behavior will be skipped
                    ret_val = ret
                    break  # Set the ret val and break

            if ret_val is None:  # If no ret_val has been set by any callback function call the supported routine
                # Emulate the routine and the return value
                ret_val = routine(self, self.pstate)

            self.__handle_external_return(routine_name, ret_val)

            # Trigger post-address callbacks
            for cb in post_cbs:
                cb(self, self.pstate, routine_name, pc)

            # Do not continue the execution if we are in a locked mutex
            if self.pstate.mutex_locked:
                self.pstate.mutex_locked = False
                self.pstate.cpu.program_counter = instruction.getAddress()
                # It's locked, so switch to another thread
                self.pstate.current_thread.count = self.config.thread_scheduling+1
                return

            # Do not continue the execution if we are in a locked semaphore
            if self.pstate.semaphore_locked:
                self.pstate.semaphore_locked = False
                self.pstate.cpu.program_counter = instruction.getAddress()
                # It's locked, so switch to another thread
                self.pstate.current_thread.count = self.config.thread_scheduling+1
                return

            if self.pstate.architecture == Architecture.AARCH64:
                # Get the return address
                ret_addr = self.pstate.read_register('x30')
            elif self.pstate.architecture in [Architecture.X86, Architecture.X86_64]:
                # Get the return address and restore RSP (simulate RET)
                ret_addr = self.pstate.pop_stack_value()
            else:
                raise Exception("Architecture not supported")

            # Hijack RIP to skip the call
            self.pstate.cpu.program_counter = ret_addr

    def _map_dynamic_symbols(self) -> None:
        """
        Apply dynamic relocations of imported functions and imported symbols
        regardless of the architecture or executable format
        .. FIXME: This function does not apply all possible relocations
        :return: None
        """
        for symbol, (addr, is_func) in self.pstate.dynamic_symbol_table.items():

            if symbol in SUPPORTED_ROUTINES:  # if the routine name is supported
                # Add link to the routine and got tables
                self.rtn_table[addr] = (symbol, SUPPORTED_ROUTINES[symbol])

            elif symbol in SUPORTED_GVARIABLES:
                # if self.pstate.architecture == Architecture.X86_64:
                self.pstate.memory.write_ptr(addr, SUPORTED_GVARIABLES[symbol])  # write directly at addr
                # elif self.pstate.architecture == Architecture.AARCH64:
                #     val = self.pstate.memory.read_ptr(addr)
                #     self.pstate.memory.write_ptr(val, SUPORTED_GVARIABLES[symbol])

            else:  # the symbol is not supported
                if self.uid == 0:  # print warning if first uid (so that it get printed once)
                    logger.warning(f"symbol {symbol} imported but unsupported")
                if is_func:
                    # Add link to a default stub function
                    self.rtn_table[addr] = (symbol, self.__default_stub)
                else:
                    pass # do nothing on unsupported symbols

    def __default_stub(self, se: 'SymbolicExecutor', pstate: ProcessState):
        rtn_name, _ = self.rtn_table[pstate.cpu.program_counter]
        logger.warning(f"calling {rtn_name} which is unsupported")
        if self.config.skip_unsupported_import:
            return None  # Like if function did nothing
        else:
            self.abort()

    def abort(self) -> NoReturn:
        """
        Abort the current execution. It works by raising
        an exception which is caught by the emulation function
        that takes care of returning appropriately afterward.

        :raise AbortExecutionException: to abort execution from anywhere
        """
        raise AbortExecutionException('Execution aborted')

    def skip_instruction(self) -> NoReturn:
        """
        Skip the current instruction before it gets executed. It is only
        relevant to call it from pre-inst or pre-addr callbacks.

        :raise SkipInstructionException: to skip the current instruction
        """
        raise SkipInstructionException("Skip instruction")

    def stop_exploration(self) -> NoReturn:
        """
        Function to call to stop the whole exploration of
        the program. It raises an exception which is caught by SymbolicExplorator.

        :raise StopExplorationException: to stop the exploration
        """
        raise StopExplorationException("Stop exploration")

    def emulation_init(self) -> bool:
        if self.pstate is None:
            logger.error(f"ProcessState is None (have you called \"load\"?")
            return False

        self.start_time = time.time()

        # Configure memory segmentation using configuration
        self.pstate.memory.set_segmentation(self.config.memory_segmentation)
        if self.config.memory_segmentation:
            self.cbm.register_memory_read_callback(self._mem_accesses_callback)
            self.cbm.register_memory_write_callback(self._mem_accesses_callback)

        # Register memory callbacks in case we activated covering mem access
        if BranchSolvingStrategy.COVER_SYM_READ in self.config.branch_solving_strategy:
            self.cbm.register_memory_read_callback(self._symbolic_mem_callback)
        if BranchSolvingStrategy.COVER_SYM_WRITE in self.config.branch_solving_strategy:
            self.cbm.register_memory_write_callback(self._symbolic_mem_callback)

        # bind dbm callbacks on the process state (newly initialized)
        self.cbm.bind_to(self)  # bind call

        # Let's emulate the binary from the entry point
        logger.info('Starting emulation')

        # Get pre/post callbacks on execution
        pre_cb, post_cb = self.cbm.get_execution_callbacks()
        # Iterate through all pre exec callbacks
        for cb in pre_cb:
            cb(self, self.pstate)

        # Call it here to make sure in case of "load_process" the use has properly instanciated the architecture
        self._configure_pstate()
        return True

    def run(self, stop_at: Addr = None) -> None:
        """
        Execute the program.

        If the :py:attr:`tritondse.Config.execution_timeout` is not set
        the execution might hang forever if the program does.

        :param stop_at: Address where to stop (if necessary)
        :return: None
        """
        if stop_at:
            self._run_to_target = stop_at

        # Call init steps
        if not self.emulation_init():
            return

        # Run until reaching a stopping condition
        self.emulate()

        # Call termination steps
        self.emulation_deinit()

    def emulation_deinit(self):
        _, post_cb = self.cbm.get_execution_callbacks()
        # Iterate through post exec callbacks
        for cb in post_cb:
            cb(self, self.pstate)

        self.end_time = time.time()

        # IMPORTANT The next call is necessary otherwise there is a memory
        #           leak issues.
        # Unbind callbacks from the current symbolic executor instance.
        self.cbm.unbind()

        # NOTE Unregister callbacks registered at the beginning of the function.
        #      This is necessary because we currently have a circular dependency
        #      between this class and the callback manager. Note that we create
        #      that circular dependency indirectly when we set the callback to
        #      a method of this class (_mem_accesses_callback and
        #      _symbolic_mem_callback).
        if self.config.memory_segmentation:
            self.cbm.unregister_callback(self._mem_accesses_callback)

        if BranchSolvingStrategy.COVER_SYM_READ in self.config.branch_solving_strategy:
            self.cbm.unregister_callback(self._symbolic_mem_callback)
        if BranchSolvingStrategy.COVER_SYM_WRITE in self.config.branch_solving_strategy:
            self.cbm.unregister_callback(self._symbolic_mem_callback)

        logger.info(f"Emulation done [ret:{self.pstate.read_register(self.pstate.return_register):x}]  (time:{self.execution_time:.02f}s)")
        logger.info(f"Instructions executed: {self.coverage.total_instruction_executed}  symbolic branches: {self.pstate.path_predicate_size}")
        logger.info(f"Memory usage: {self.mem_usage_str()}")

    def _mem_accesses_callback(self, se: 'SymbolicExecutor', ps: ProcessState, mem: MemoryAccess, *args):
        """
        This callback is used to ensure memory accesses performed by side-effect of instructions
        semantic correctly checks memory segmentation. Thus we only do the check during the processing
        of an instruction.
        """
        if ps.memory.segmentation_enabled and self._in_processing:
            perm = Perm.W if bool(args) else Perm.R
            addr = mem.getAddress()
            size = mem.getSize()
            map = ps.memory.get_map(addr, size)  # It raises
            if map is None:
                raise MemoryAccessViolation(addr, perm, memory_not_mapped=True)
            else:
                if perm not in map.perm:
                    raise MemoryAccessViolation(addr, perm, map_perm=map.perm, perm_error=True)

    @property
    def exitcode(self) -> int:
        """ Exit code value of the process. The value
        is simply the concrete value of the register
        marked as return_register (rax, on x86, r0 on ARM..)
        """
        return self.pstate.read_register(self.pstate.return_register) & 0xFF

    @staticmethod
    def mem_usage_str() -> str:
        """
        Debug function to track memory consumption of an execution (not
        implemented on Windows).
        """
        if os.name == "posix":
            size, resident, shared, _, _, _, _ = (int(x) for x in open(f"/proc/{os.getpid()}/statm").read().split(" "))
            resident = resident * resource.getpagesize()
            units = [(float(1024), "Kb"), (float(1024 **2), "Mb"), (float(1024 **3), "Gb")]
            for unit, s in units[::-1]:
                if resident / unit < 1:
                    continue
                else:  # We are on the right unit
                    return "%.2f%s" % (resident/unit, s)
            return "%dB" % resident
        else:
          return "N/A"

    def mk_new_seed_from_model(self, model: Model) -> Seed:
        """
        Creates a new seed from the given SMT model.

        :param model: SMT model
        :return: new seed object
        """
        def repl_bytearray(concrete, symbolic):
            for i, sv in enumerate(symbolic):  # Enumerate symvars associated with each bytes
                if sv is not None:
                    if sv.getId() in model:  # If solver provided a new value for the symvar
                        value = model[sv.getId()].getValue()
                        concrete[i] = value # Replace it in the bytearray
            return concrete

        if self.config.is_format_raw(): # RAW seed. => symbolize_stdin
            content = bytes(repl_bytearray(bytearray(self.seed.content), self.symbolic_seed))

        elif self.config.is_format_composite():
            # NOTE will have to update this if more things are added to CompositeData
            new_files, new_vars = {}, {}

            # Handle argv (its meant to be here)
            args = [bytearray(x) for x in self.seed.content.argv]
            new_argv = [bytes(repl_bytearray(c, s)) for c, s in zip(args, self.symbolic_seed.argv)]

            # Handle stdin and files
            # If the seed provides the content of files (#NOTE stdin is treated as a file)
            new_files = {}
            for k, c in self.seed.content.files.items():
                if k in self.symbolic_seed.files:
                    new_files[k] = bytes(repl_bytearray(bytearray(c), self.symbolic_seed.files[k]))
                else:
                    new_files[k] = c  # keep the current value in the seed

            # Handle variables, if the seed provides some
            new_variables = {}
            for k, c in self.seed.content.variables.items():
                if k in self.symbolic_seed.variables:
                    conc = bytearray(c) if isinstance(c, bytes) else [c]
                    new_vals = repl_bytearray(conc, self.symbolic_seed.variables[k])
                    new_variables[k] = bytes(new_vals) if isinstance(c, bytes) else new_vals[0]  # new variables are either bytes or int
                else:
                    new_variables[k] = c  # If it has not been injected keep the current concrete value

            content = CompositeData(new_argv, new_files, new_variables)
        else:
            assert False

        # Calling callback if user defined one
        new_seed = Seed(content)
        for cb in self.cbm.get_new_input_callback():
            cont = cb(self, self.pstate, new_seed)
            if cont:
                # if the callback return a new input continue with that one
                new_seed = cont
        # Return the
        return new_seed

    def inject_symbolic_argv_memory(self, addr: Addr, index: int, value: bytes) -> None:
        """
        Inject the ith item of argv in memory.
        To be used only with composite seeds and only if seed have a symbolic argv

        :param addr: address where to inject the argv[ith]
        :param index: ith argv item
        :param value: value of the item
        """
        self.pstate.memory.write(addr, value)  # Write concrete bytes in memory
        sym_vars = self.pstate.symbolize_memory_bytes(addr, len(value), f"argv[{index}]") # Symbolize bytes
        self.symbolic_seed.argv[index] = sym_vars # Add symbolic variables to symbolic seed

    def inject_symbolic_file_memory(self, addr: Addr, name: str, value: bytes, offset: int = 0) -> None:
        """
        Inject a symbolic file (or part of it) in memory.

        :param addr: address where to inject the file bytes
        :param name: name of the file in the composite seed
        :param value: bytes content of the file
        :param offset: offset within the file (for partial file injection)
        """
        self.pstate.memory.write(addr, value)  # Write concrete bytes in memory
        sym_vars = self.pstate.symbolize_memory_bytes(addr, len(value), name, offset) # Symbolize bytes
        sym_seed = self.symbolic_seed.files[name] if self.seed.is_composite() else self.symbolic_seed
        sym_seed[offset:offset+len(value)] = sym_vars # Add symbolic variables to symbolic seed
        # FIXME: Handle if reading twice same input bytes !

    def inject_symbolic_variable_memory(self, addr: Addr, name: str, value: bytes, offset: int = 0) -> None:
        """
        Inject a symbolic variable in memory.

        :param addr: address where to inject the variable
        :param name: name of the variable in the composite seed
        :param value: value of the variable
        :param offset: offset within the variable (for partial variable injection)
        :return:
        """
        self.pstate.memory.write(addr, value)  # Write concrete bytes in memory
        sym_vars = self.pstate.symbolize_memory_bytes(addr, len(value), name, offset)  # Symbolize bytes
        self.symbolic_seed.variables[name][offset:offset+len(value)-1] = sym_vars # Add symbolic variables to symbolic seed
        # FIXME: Handle if reading twice same input bytes !

    def inject_symbolic_file_register(self, reg: Union[str, Register], name: str, value: int, offset: int = 0) -> None:
        """
        Inject a symbolic file (or part of it) into a register.
        The value has to be an integer.

        :param reg: register identifier
        :param name: name of the file in the composite seed
        :param value: integer value
        :param offset: offset within the file
        """
        if reg.getSize != 1:
            logger.error("can't call inject_symbolic_file_register with regsiter larger than 1!")
            return
        self.pstate.write_register(reg, value)  # Write concrete value in register
        sym_vars = self.pstate.symbolize_register(reg, f"{name}[{offset}]")  # Symbolize bytes
        sym_seed = self.symbolic_seed.files[name] if self.seed.is_composite() else self.symbolic_seed
        sym_seed[offset] = sym_vars  # Add symbolic variables to symbolic seed

    def inject_symbolic_variable_register(self, reg: Union[str, Register], name: str, value: int) -> None:
        """
        Inject a symbolic variable (or part of it) in a register.
        The value has to be an integer.

        :param reg: register identifier
        :param name: name of the variable
        :param value: integer value
        """
        if not self.seed.is_composite():
            logger.warning("cannot use inject_symbolic_variable_register on raw seeds!")
            return

        if isinstance(value, int):
            self.pstate.write_register(reg, value)                         # write concrete value in register
            sym_var = self.pstate.symbolize_register(reg, f"{name}[{0}]")  # symbolize value
            self.symbolic_seed.variables[name][0] = sym_var               # add the symbolic variables to symbolic seed
        else:  # meant to be bytes
            logger.warning("variable injected in registers have to be integer values")

    def inject_symbolic_raw_input(self, addr: Addr, data: bytes, offset: int = 0) -> None:
        """
        Inject the input in memory. This injection method should
        be used for RAW seed type.

        :param addr: address where to inject input.
        :param data: content of the seed
        :param offset: offset within the content of the seed.
        """
        if self.seed.is_composite():
            logger.warning("inject_symbolic_memory must not be used with composite seeds !")
        else:
            self.inject_symbolic_file_memory(addr, "input", data, offset)



import io
import logging
import os
import random
import re
import sys
import time

from tritondse.types import Architecture, FileDesc
from tritondse.seed import SeedFormat, SeedStatus, Seed
import tritondse.logging

logger = tritondse.logging.get("routines")

NULL_PTR = 0



def rtn_ctype_b_loc(se: 'SymbolicExecutor', pstate: 'ProcessState'):
    """
    The __ctype_b_loc behavior.
    """
    logger.debug('__ctype_b_loc hooked')

    ctype  = b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    ctype += b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    ctype += b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    ctype += b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    ctype += b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    ctype += b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    ctype += b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    ctype += b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    ctype += b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    ctype += b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    ctype += b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    ctype += b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    ctype += b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    ctype += b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    ctype += b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    ctype += b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    ctype += b"\x02\x00\x02\x00\x02\x00\x02\x00\x02\x00\x02\x00\x02\x00\x02\x00"  # must point here
    ctype += b"\x02\x00\x03\x20\x02\x20\x02\x20\x02\x20\x02\x20\x02\x00\x02\x00"
    ctype += b"\x02\x00\x02\x00\x02\x00\x02\x00\x02\x00\x02\x00\x02\x00\x02\x00"
    ctype += b"\x02\x00\x02\x00\x02\x00\x02\x00\x02\x00\x02\x00\x02\x00\x02\x00"
    ctype += b"\x01\x60\x04\xc0\x04\xc0\x04\xc0\x04\xc0\x04\xc0\x04\xc0\x04\xc0"
    ctype += b"\x04\xc0\x04\xc0\x04\xc0\x04\xc0\x04\xc0\x04\xc0\x04\xc0\x04\xc0"
    ctype += b"\x08\xd8\x08\xd8\x08\xd8\x08\xd8\x08\xd8\x08\xd8\x08\xd8\x08\xd8"
    ctype += b"\x08\xd8\x08\xd8\x04\xc0\x04\xc0\x04\xc0\x04\xc0\x04\xc0\x04\xc0"
    ctype += b"\x04\xc0\x08\xd5\x08\xd5\x08\xd5\x08\xd5\x08\xd5\x08\xd5\x08\xc5"
    ctype += b"\x08\xc5\x08\xc5\x08\xc5\x08\xc5\x08\xc5\x08\xc5\x08\xc5\x08\xc5"
    ctype += b"\x08\xc5\x08\xc5\x08\xc5\x08\xc5\x08\xc5\x08\xc5\x08\xc5\x08\xc5"
    ctype += b"\x08\xc5\x08\xc5\x08\xc5\x04\xc0\x04\xc0\x04\xc0\x04\xc0\x04\xc0"
    ctype += b"\x04\xc0\x08\xd6\x08\xd6\x08\xd6\x08\xd6\x08\xd6\x08\xd6\x08\xc6"
    ctype += b"\x08\xc6\x08\xc6\x08\xc6\x08\xc6\x08\xc6\x08\xc6\x08\xc6\x08\xc6"
    ctype += b"\x08\xc6\x08\xc6\x08\xc6\x08\xc6\x08\xc6\x08\xc6\x08\xc6\x08\xc6"
    ctype += b"\x08\xc6\x08\xc6\x08\xc6\x04\xc0\x04\xc0\x04\xc0\x04\xc0\x02\x00"
    ctype += b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    ctype += b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    ctype += b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    ctype += b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    ctype += b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    ctype += b"\x00\x00\x00\x00\x00\x00\x00\x00"

    # Allocate on heap enough to make the table to fit
    alloc_size = 2*pstate.ptr_size + len(ctype)
    base_ctype = pstate.heap_allocator.alloc(alloc_size)

    ctype_table_offset = base_ctype + (pstate.ptr_size * 2)
    otable_offset = ctype_table_offset + 256

    pstate.memory.write_ptr(base_ctype, otable_offset)
    pstate.memory.write_ptr(base_ctype+pstate.ptr_size, 0)

    # FIXME: On pourrait la renvoyer qu'une seule fois ou la charger au demarage direct dans pstate
    pstate.memory.write(ctype_table_offset, ctype)

    return base_ctype

def rtn_ctype_toupper_loc(se: 'SymbolicExecutor', pstate: 'ProcessState'):
    # FIXME: Not sure about the array and where to place the pointer 
    # https://codebrowser.dev/glibc/glibc/locale/C-ctype.c.html
    """
    The __ctype_toupper_loc behavior.
    """
    logger.debug('__ctype_toupper_loc hooked')

    ctype  = b"\x80\x81\x82\x83\x84\x85\x86\x87\x88\x89\x8a\x8b\x8c\x8d\x8e\x8f"
    ctype += b"\x90\x91\x92\x93\x94\x95\x96\x97\x98\x99\x9a\x9b\x9c\x9d\x9e\x9f"
    ctype += b"\xa0\xa1\xa2\xa3\xa4\xa5\xa6\xa7\xa8\xa9\xaa\xab\xac\xad\xae\xaf"
    ctype += b"\xb0\xb1\xb2\xb3\xb4\xb5\xb6\xb7\xb8\xb9\xba\xbb\xbc\xbd\xbe\xbf"
    ctype += b"\xc0\xc1\xc2\xc3\xc4\xc5\xc6\xc7\xc8\xc9\xca\xcb\xcc\xcd\xce\xcf"
    ctype += b"\xd0\xd1\xd2\xd3\xd4\xd5\xd6\xd7\xd8\xd9\xda\xdb\xdc\xdd\xde\xdf"
    ctype += b"\xe0\xe1\xe2\xe3\xe4\xe5\xe6\xe7\xe8\xe9\xea\xeb\xec\xed\xee\xef"
    ctype += b"\xf0\xf1\xf2\xf3\xf4\xf5\xf6\xf7\xf8\xf9\xfa\xfb\xfc\xfd\xfe\xff\xff\xff\xff"
    ctype += b"\x00\x01\x02\x03\x04\x05\x06\x07\x08\x09\x0a\x0b\x0c\x0d\x0e\x0f"
    ctype += b"\x10\x11\x12\x13\x14\x15\x16\x17\x18\x19\x1a\x1b\x1c\x1d\x1e\x1f"
    ctype += b"\x20\x21\x22\x23\x24\x25\x26\x27\x28\x29\x2a\x2b\x2c\x2d\x2e\x2f"
    ctype += b"\x30\x31\x32\x33\x34\x35\x36\x37\x38\x39\x3a\x3b\x3c\x3d\x3e\x3f"
    ctype += b"\x40\x41\x42\x43\x44\x45\x46\x47\x48\x49\x4a\x4b\x4c\x4d\x4e\x4f"
    ctype += b"\x50\x51\x52\x53\x54\x55\x56\x57\x58\x59\x5a\x5b\x5c\x5d\x5e\x5f"
    ctype += b"\x60\x41\x42\x43\x44\x45\x46\x47\x48\x49\x4a\x4b\x4c\x4d\x4e\x4f"
    ctype += b"\x50\x51\x52\x53\x54\x55\x56\x57\x58\x59\x5a\x7b\x7c\x7d\x7e\x7f"
    ctype += b"\x80\x81\x82\x83\x84\x85\x86\x87\x88\x89\x8a\x8b\x8c\x8d\x8e\x8f"
    ctype += b"\x90\x91\x92\x93\x94\x95\x96\x97\x98\x99\x9a\x9b\x9c\x9d\x9e\x9f"
    ctype += b"\xa0\xa1\xa2\xa3\xa4\xa5\xa6\xa7\xa8\xa9\xaa\xab\xac\xad\xae\xaf"
    ctype += b"\xb0\xb1\xb2\xb3\xb4\xb5\xb6\xb7\xb8\xb9\xba\xbb\xbc\xbd\xbe\xbf"
    ctype += b"\xc0\xc1\xc2\xc3\xc4\xc5\xc6\xc7\xc8\xc9\xca\xcb\xcc\xcd\xce\xcf"
    ctype += b"\xd0\xd1\xd2\xd3\xd4\xd5\xd6\xd7\xd8\xd9\xda\xdb\xdc\xdd\xde\xdf"
    ctype += b"\xe0\xe1\xe2\xe3\xe4\xe5\xe6\xe7\xe8\xe9\xea\xeb\xec\xed\xee\xef"
    ctype += b"\xf0\xf1\xf2\xf3\xf4\xf5\xf6\xf7\xf8\xf9\xfa\xfb\xfc\xfd\xfe\xff"

    # Allocate on heap enough to make the table to fit
    alloc_size = 2*pstate.ptr_size + len(ctype)
    base_ctype = pstate.heap_allocator.alloc(alloc_size)

    ctype_table_offset = base_ctype + (pstate.ptr_size * 2)
    otable_offset = ctype_table_offset + 256

    pstate.memory.write_ptr(base_ctype, otable_offset)
    pstate.memory.write_ptr(base_ctype+pstate.ptr_size, 0)

    # FIXME: On pourrait la renvoyer qu'une seule fois ou la charger au demarage direct dans pstate
    pstate.memory.write(ctype_table_offset, ctype)

    return base_ctype


def rtn_errno_location(se: 'SymbolicExecutor', pstate: 'ProcessState'):
    """
    The __errno_location behavior.
    """
    logger.debug('__errno_location hooked')

    # Errno is a int* ptr, initialize it to zero
    # We consider it is located in the [extern] segment
    # Thus the process must have one of this map
    segs = pstate.memory.find_map(pstate.EXTERN_SEG)
    if segs:
        map = segs[0]
        ERRNO = map.start + map.size - 4  # Point is last int of the mapping
    else:
        assert False
    pstate.memory.write_dword(ERRNO, 0)

    return ERRNO


def rtn_libc_start_main(se: 'SymbolicExecutor', pstate: 'ProcessState'):
    """
    The __libc_start_main behavior.
    """
    logger.debug('__libc_start_main hooked')

    # Get arguments
    main = pstate.get_argument_value(0)

    # WARNING: Dirty trick to make sure to jump on main after
    # the emulation of that stub
    if pstate.architecture == Architecture.AARCH64:
        pstate.cpu.x30 = main
    elif pstate.architecture in [Architecture.X86_64, Architecture.X86]:
        # Push the return value to jump into the main() function
        pstate.push_stack_value(main)
    else:
        assert False

    # Define concrete value of argc (from either the seed or the program_argv)
    if se.config.is_format_raw():
        # Cannot provide argv in RAW seeds
        argc = len(se.config.program_argv)
    else: # SeedFormat.COMPOSITE
        argc = len(se.seed.content.argv) if se.seed.content.argv else len(se.config.program_argv)

    if pstate.architecture == Architecture.X86:
        # Because of the "Dirty trick" described above, we RET to main instead of CALLing it.
        # Because of that, the arguments end up 1 slot off on the stack
        pstate.write_argument_value(0 + 1, argc)
    else:
        pstate.write_argument_value(0, argc)
    logger.debug(f"argc = {argc}")

    # Define argv
    addrs = list()

    if se.config.is_format_composite() and se.seed.content.argv: # Use the seed provided (and ignore config.program_argv !!)
        argvs = se.seed.content.argv
        src = 'seed'
    else:  # use the config argv
        argvs = [x.encode("latin-1") for x in se.config.program_argv]  # Convert it from str to bytes
        src = 'config'

    # Compute the allocation size: size of strings, + all \x00 + all pointers
    size = sum(len(x) for x in argvs)+len(argvs)+len(argvs)*pstate.ptr_size
    if size == 0:  # Falback on a single pointer that will hold not even be initialized
        size = pstate.ptr_size

    # We put the ARGV stuff on the heap even though its normally on stack
    base = pstate.heap_allocator.alloc(size)

    for i, arg in enumerate(argvs):
        addrs.append(base)
        pstate.memory.write(base, arg + b'\x00')

        if se.config.is_format_composite() and se.seed.content.argv: # Use the seed provided (and ignore config.program_argv !!)
            # Symbolize the argv string
            se.inject_symbolic_argv_memory(base, i, arg)
            # FIXME: Shall add a constraint on every char to be != \x00

        logger.debug(f"({src}) argv[{i}] = {repr(pstate.memory.read(base, len(arg)))}")
        base += len(arg) + 1


    # NOTE: the array of pointers will be after the string themselves
    b_argv = base
    for addr in addrs:
        pstate.memory.write_ptr(base, addr)
        base += pstate.ptr_size

    # Concrete value
    if pstate.architecture == Architecture.X86:
        # Because of the "Dirty trick" described above, we RET to main instead of CALLing it.
        # Because of that, the arguments end up 1 slot off on the stack
        pstate.write_argument_value(1 + 1, b_argv)
    else:
        pstate.write_argument_value(1, b_argv)
    return None


def rtn_stack_chk_fail(se: 'SymbolicExecutor', pstate: 'ProcessState'):
    """
    The __stack_chk_fail behavior.
    """
    logger.debug('__stack_chk_fail hooked')
    logger.critical('*** stack smashing detected ***: terminated')
    se.seed.status = SeedStatus.CRASH
    pstate.stop = True


# int __xstat(int ver, const char* path, struct stat* stat_buf);
def rtn_xstat(se: 'SymbolicExecutor', pstate: 'ProcessState'):
    """
    The __xstat behavior.
    """
    logger.debug('__xstat hooked')

    # Get arguments
    arg0 = pstate.get_argument_value(0)  # int ver
    arg1 = pstate.get_argument_value(1)  # const char* path
    arg2 = pstate.get_argument_value(2)  # struct stat* stat_buf

    if os.path.isfile(pstate.memory.read_string(arg1)):
        stat = os.stat(pstate.memory.read_string(arg1))
        pstate.memory.write_qword(arg2 + 0x00, stat.st_dev)
        pstate.memory.write_qword(arg2 + 0x08, stat.st_ino)
        pstate.memory.write_qword(arg2 + 0x10, stat.st_nlink)
        pstate.memory.write_dword(arg2 + 0x18, stat.st_mode)
        pstate.memory.write_dword(arg2 + 0x1c, stat.st_uid)
        pstate.memory.write_dword(arg2 + 0x20, stat.st_gid)
        pstate.memory.write_dword(arg2 + 0x24, 0)
        pstate.memory.write_qword(arg2 + 0x28, stat.st_rdev)
        pstate.memory.write_qword(arg2 + 0x30, stat.st_size)
        pstate.memory.write_qword(arg2 + 0x38, stat.st_blksize)
        pstate.memory.write_qword(arg2 + 0x40, stat.st_blocks)
        pstate.memory.write_qword(arg2 + 0x48, 0)
        pstate.memory.write_qword(arg2 + 0x50, 0)
        pstate.memory.write_qword(arg2 + 0x58, 0)
        pstate.memory.write_qword(arg2 + 0x60, 0)
        pstate.memory.write_qword(arg2 + 0x68, 0)
        pstate.memory.write_qword(arg2 + 0x70, 0)
        pstate.memory.write_qword(arg2 + 0x78, 0)
        pstate.memory.write_qword(arg2 + 0x80, 0)
        pstate.memory.write_qword(arg2 + 0x88, 0)
        return 0

    return pstate.minus_one


def rtn_abort(se: 'SymbolicExecutor', pstate: 'ProcessState'):
    """::

        void abort(void);

    Mark the input seed as OK and stop execution.

    [`Man Page <https://man7.org/linux/man-pages/man3/abort.3.html>`_]
    """
    logger.debug('abort hooked')
    se.seed.status = SeedStatus.OK_DONE
    pstate.stop = True


# int atoi(const char *nptr);
def rtn_atoi(se: 'SymbolicExecutor', pstate: 'ProcessState'):
    """::

        int atoi(const char *nptr);

    **Description**: The atoi() function converts the initial portion of the string
    pointed to by nptr to int.  The behavior is the same as

    Concrete: /

    Symbolic: Represent the return value symbolically with 10 nested if
    to represent the value.

    .. warning:: The function does not support all possibles representation
                 of an integer. It does not support negative integer nor values
                 prefixed by spaces.

    [`Man Page <https://man7.org/linux/man-pages/man3/abort.3.html>`_]

    :return: Symbolic value of the integer base on the symbolic string ``nptr``
    """
    logger.debug('atoi hooked')

    ast = pstate.actx
    arg = pstate.get_argument_value(0)

    # FIXME: On ne concretize pas correctement la taille de la chaine

    cells = {i: pstate.read_symbolic_memory_byte(arg+i).getAst() for i in range(10)}

    # FIXME: Does not support negative value and all other corner cases.
    # "000000000012372183762173"
    # "         98273483274"
    # "-123123"
    # "18273213"

    def multiply(ast, cells, index):
        n = ast.bv(0, 32)
        for i in range(index):
            n = n * 10 + (ast.zx(24, cells[i]) - 0x30)
        return n

    res = ast.ite(
              ast.lnot(ast.land([cells[0] >= 0x30, cells[0] <= 0x39])),
              multiply(ast, cells, 0),
              ast.ite(
                  ast.lnot(ast.land([cells[1] >= 0x30, cells[1] <= 0x39])),
                  multiply(ast, cells, 1),
                  ast.ite(
                      ast.lnot(ast.land([cells[2] >= 0x30, cells[2] <= 0x39])),
                      multiply(ast, cells, 2),
                      ast.ite(
                          ast.lnot(ast.land([cells[3] >= 0x30, cells[3] <= 0x39])),
                          multiply(ast, cells, 3),
                          ast.ite(
                              ast.lnot(ast.land([cells[4] >= 0x30, cells[4] <= 0x39])),
                              multiply(ast, cells, 4),
                              ast.ite(
                                  ast.lnot(ast.land([cells[5] >= 0x30, cells[5] <= 0x39])),
                                  multiply(ast, cells, 5),
                                  ast.ite(
                                      ast.lnot(ast.land([cells[6] >= 0x30, cells[6] <= 0x39])),
                                      multiply(ast, cells, 6),
                                      ast.ite(
                                          ast.lnot(ast.land([cells[7] >= 0x30, cells[7] <= 0x39])),
                                          multiply(ast, cells, 7),
                                          ast.ite(
                                              ast.lnot(ast.land([cells[8] >= 0x30, cells[8] <= 0x39])),
                                              multiply(ast, cells, 8),
                                              ast.ite(
                                                  ast.lnot(ast.land([cells[9] >= 0x30, cells[9] <= 0x39])),
                                                  multiply(ast, cells, 9),
                                                  multiply(ast, cells, 9)
                                              )
                                          )
                                      )
                                  )
                              )
                          )
                      )
                  )
              )
          )
    res = ast.sx(32, res)

    return res


# void *calloc(size_t nmemb, size_t size);
def rtn_calloc(se: 'SymbolicExecutor', pstate: 'ProcessState'):
    """
    The calloc behavior.
    """
    logger.debug('calloc hooked')

    # Get arguments
    nmemb = pstate.get_argument_value(0)
    size  = pstate.get_argument_value(1)

    # We use nmemb and size as concret values
    pstate.concretize_argument(0)  # will be concretized with nmemb value
    pstate.concretize_argument(1)  # will be concretized with size value

    if nmemb == 0 or size == 0:
        ptr = NULL_PTR
    else:
        ptr = pstate.heap_allocator.alloc(nmemb * size)
        # Once the ptr allocated, the memory area must be filled with zero
        for index in range(nmemb * size):
            pstate.write_symbolic_memory_byte(ptr+index, pstate.actx.bv(0, 8))

    # Return value
    return ptr


# int clock_gettime(clockid_t clockid, struct timespec *tp);
def rtn_clock_gettime(se: 'SymbolicExecutor', pstate: 'ProcessState'):
    """
    The clock_gettime behavior.
    """
    logger.debug('clock_gettime hooked')

    # Get arguments
    clockid = pstate.get_argument_value(0)
    tp      = pstate.get_argument_value(1)

    # We use tp as concret value
    pstate.concretize_argument(1)

    # FIXME: We can return something logic
    if tp == 0:
        return pstate.minus_one

    if pstate.time_inc_coefficient:
        t = pstate.time
    else:
        t = time.time()

    pstate.memory.write_ptr(tp, int(t))
    pstate.memory.write_ptr(tp+pstate.ptr_size, int(t * 1000000))

    # Return value
    return 0


def rtn_exit(se: 'SymbolicExecutor', pstate: 'ProcessState'):
    """
    The exit behavior.
    """
    logger.debug('exit hooked')
    arg = pstate.get_argument_value(0)
    pstate.stop = True
    return arg


def rtn_fclose(se: 'SymbolicExecutor', pstate: 'ProcessState'):
    """
    The fclose behavior.
    """
    logger.debug('fclose hooked')

    # Get arguments
    arg0 = pstate.get_argument_value(0) # fd

    # We use fd as concret value
    pstate.concretize_argument(0)

    if pstate.file_descriptor_exists(arg0):
        pstate.close_file_descriptor(arg0)
    else:
        return pstate.minus_one

    # Return value
    return 0


def rtn_fseek(se: 'SymbolicExecutor', pstate: 'ProcessState'):
    """
    The fseek behavior.
    """

    #define SEEK_SET    0   /* set file offset to offset */
    #define SEEK_CUR    1   /* set file offset to current plus offset */
    #define SEEK_END    2   /* set file offset to EOF plus offset */
    logger.debug('fseek hooked')

    # Get arguments
    arg0 = pstate.get_argument_value(0)
    arg1 = pstate.get_argument_value(1)
    arg2 = pstate.get_argument_value(2)

    if arg2 not in [0, 1, 2]:
        return pstate.minus_one
        # TODO: set errno to: -22 # EINVAL

    if pstate.file_descriptor_exists(arg0):
        desc = pstate.get_file_descriptor(arg0)

        if desc.fd.seekable():
            r = desc.fd.seek(arg1, arg2)
            return r
        else:
            return -1
            # TODO: set errno to: 29 # ESPIPE
    else:
        return -1


def rtn_ftell(se: 'SymbolicExecutor', pstate: 'ProcessState'):
    """
    The ftell behavior.
    """

    logger.debug('ftell hooked')

    # Get arguments
    arg0 = pstate.get_argument_value(0)

    if pstate.file_descriptor_exists(arg0):
        desc = pstate.get_file_descriptor(arg0)

        if desc.fd.seekable():
            return desc.fd.tell()
        else:
            return -1
            # TODO: set errno -22 EINVAL


# char *fgets(char *s, int size, FILE *stream);
def rtn_fgets(se: 'SymbolicExecutor', pstate: 'ProcessState'):
    """
    The fgets behavior.
    """
    logger.debug('fgets hooked')

    # Get arguments
    buff, buff_ast = pstate.get_full_argument(0)
    size, size_ast = pstate.get_full_argument(1)
    fd = pstate.get_argument_value(2)
    # We use fd as concret value
    pstate.concretize_argument(2)

    if pstate.file_descriptor_exists(fd):
        filedesc = pstate.get_file_descriptor(fd)
        offset = filedesc.offset
        data = filedesc.fgets(size)
        data_no_trail = data[:-1]  # not \x00 terminated

        if filedesc.is_input_fd(): # Reading into input
            # if we started from empty seed simulate reading `size` amount of data
            if se.seed.is_raw() and se.seed.is_bootstrap_seed() and not data_no_trail:
                data = b'\x00' * size

            if len(data_no_trail) == size:  # if `size` limited the fgets its an indirect constraint
                pstate.push_constraint(size_ast.getAst() == size)

            se.inject_symbolic_file_memory(buff, filedesc.name, data, offset)
            logger.debug(f"fgets() in {filedesc.name} = {repr(data)}")
        else:
            pstate.concretize_argument(1)
            pstate.memory.write(buff, data)

        return buff_ast if data_no_trail else NULL_PTR
    else:
        logger.warning(f'File descriptor ({fd}) not found')
        return NULL_PTR


# fopen(const char *pathname, const char *mode);
def rtn_fopen(se: 'SymbolicExecutor', pstate: 'ProcessState'):
    """
    The fopen behavior.
    """
    logger.debug('fopen hooked')

    # Get arguments
    arg0 = pstate.get_argument_value(0)  # const char *pathname
    arg1 = pstate.get_argument_value(1)  # const char *mode
    arg0s = pstate.memory.read_string(arg0)
    arg1s = pstate.memory.read_string(arg1)

    # Concretize the whole path name
    pstate.concretize_memory_bytes(arg0, len(arg0s)+1)  # Concretize the whole string + \0

    # We use mode as concrete value
    pstate.concretize_argument(1)

    if se.seed.is_file_defined(arg0s):
        logger.info(f"opening an input file: {arg0s}")
        # Program is opening an input
        data = se.seed.get_file_input(arg0s)
        filedesc = pstate.create_file_descriptor(arg0s, io.BytesIO(data))
        return filedesc.id
    else:
        # Try to open it as a regular file
        try:
            fd = open(arg0s, arg1s)
            filedesc = pstate.create_file_descriptor(arg0s, fd)
            return filedesc.id
        except Exception as e:
            logger.debug(f"Failed to open {arg0s} {e}")
            return NULL_PTR


def rtn_fprintf(se: 'SymbolicExecutor', pstate: 'ProcessState'):
    """
    The fprintf behavior.
    """
    logger.debug('fprintf hooked')

    # Get arguments
    arg0 = pstate.get_argument_value(0)
    arg1 = pstate.get_argument_value(1)

    # FIXME: ARM64
    # FIXME: pushPathConstraint

    arg1f = pstate.get_format_string(arg1)
    nbArgs = arg1f.count("{")
    args = pstate.get_format_arguments(arg1, [pstate.get_argument_value(x) for x in range(2, nbArgs+2)])
    try:
        s = arg1f.format(*args)
    except:
        # FIXME: Les chars UTF8 peuvent foutre le bordel. Voir avec ground-truth/07.input
        logger.warning('Something wrong, probably UTF-8 string')
        s = ""

    if pstate.file_descriptor_exists(arg0):
        fdesc = pstate.get_file_descriptor(arg0)
        if arg0 not in [1, 2] or (arg0 == 1 and se.config.pipe_stdout) or (arg0 == 2 and se.config.pipe_stderr):
            fdesc.fd.write(s)
            fdesc.fd.flush()
    else:
        return 0

    # Return value
    return len(s)


def rtn___fprintf_chk(se: 'SymbolicExecutor', pstate: 'ProcessState'):
    """
    The __fprintf_chk behavior.
    """
    logger.debug('__fprintf_chk hooked')

    # Get arguments
    arg0 = pstate.get_argument_value(0)
    flag = pstate.get_argument_value(1)
    arg1 = pstate.get_argument_value(2)

    # FIXME: ARM64
    # FIXME: pushPathConstraint

    arg1f = pstate.get_format_string(arg1)
    nbArgs = arg1f.count("{")
    args = pstate.get_format_arguments(arg1, [pstate.get_argument_value(x) for x in range(3, nbArgs+2)])
    try:
        s = arg1f.format(*args)
    except:
        # FIXME: Les chars UTF8 peuvent foutre le bordel. Voir avec ground-truth/07.input
        logger.warning('Something wrong, probably UTF-8 string')
        s = ""

    if pstate.file_descriptor_exists(arg0):
        fdesc = pstate.get_file_descriptor(arg0)
        if arg0 not in [1, 2] or (arg0 == 1 and se.config.pipe_stdout) or (arg0 == 2 and se.config.pipe_stderr):
            fdesc.fd.write(s)
            fdesc.fd.flush()
    else:
        return 0

    # Return value
    return len(s)


# fputc(int c, FILE *stream);
def rtn_fputc(se: 'SymbolicExecutor', pstate: 'ProcessState'):
    """
    The fputc behavior.
    """
    logger.debug('fputc hooked')

    # Get arguments
    arg0 = pstate.get_argument_value(0)
    arg1 = pstate.get_argument_value(1)

    pstate.concretize_argument(0)
    pstate.concretize_argument(1)

    if pstate.file_descriptor_exists(arg1):
        fdesc = pstate.get_file_descriptor(arg1)
        if arg1 == 0:
            return 0
        elif (arg1 == 1 and se.config.pipe_stdout) or (arg1 == 2 and se.config.pipe_stderr):
            fdesc.fd.write(chr(arg0))
            fdesc.fd.flush()
        elif arg1 not in [0, 2]:
            fdesc.fd.write(chr(arg0))
        return 1
    else:
        return 0



def rtn_fputs(se: 'SymbolicExecutor', pstate: 'ProcessState'):
    """
    The fputs behavior.
    """
    logger.debug('fputs hooked')

    # Get arguments
    arg0 = pstate.get_argument_value(0)
    arg1 = pstate.get_argument_value(1)

    pstate.concretize_argument(0)
    pstate.concretize_argument(1)

    s = pstate.memory.read_string(arg0)

    # FIXME: What if the fd is coming from the memory (fmemopen) ?

    if pstate.file_descriptor_exists(arg1):
        fdesc = pstate.get_file_descriptor(arg1)
        if arg1 == 0:
            return 0
        elif arg1 == 1:
            if se.config.pipe_stdout:
                fdesc.fd.write(s)
                fdesc.fd.flush()
        elif arg1 == 2:
            if se.config.pipe_stderr:
                fdesc.fd.write(s)
                fdesc.fd.flush()
        else:
            fdesc.fd.write(s)
    else:
        return 0

    # Return value
    return len(s)


def rtn_fread(se: 'SymbolicExecutor', pstate: 'ProcessState'):
    """
    The fread behavior.
    """
    logger.debug('fread hooked')

    # Get arguments
    ptr = pstate.get_argument_value(0) # ptr
    size_t, size_ast = pstate.get_full_argument(1) # size
    nmemb = pstate.get_argument_value(2) # nmemb
    fd = pstate.get_argument_value(3) # stream
    size = size_t * nmemb

    # FIXME: pushPathConstraint

    if pstate.file_descriptor_exists(fd):
        filedesc = pstate.get_file_descriptor(fd)
        offset = filedesc.offset
        data = filedesc.read(size)

        if filedesc.is_input_fd(): # Reading into input
            # if we started from empty seed simulate reading `size` amount of data
            if se.seed.is_raw() and se.seed.is_bootstrap_seed() and not data:
                data = b'\x00' * size

            if len(data) == size:  # if `size` limited the fgets its an indirect constraint
                pstate.push_constraint(size_ast.getAst() == size)

            se.inject_symbolic_file_memory(ptr, filedesc.name, data, offset)
            logger.debug(f"read in {filedesc.name} = {repr(data)}")
        else:
            pstate.concretize_argument(2)
            pstate.memory.write(ptr, data)

        return int(len(data)/size_t) if size_t else 0  # number of items read
    else:
        logger.warning(f'File descriptor ({fd}) not found')
        return 0


def rtn_free(se: 'SymbolicExecutor', pstate: 'ProcessState'):
    """
    The free behavior.
    """
    logger.debug('free hooked')

    # Get arguments
    ptr = pstate.get_argument_value(0)
    if ptr == 0: # free(NULL) is a nop
        return None
    pstate.heap_allocator.free(ptr)

    return None


def rtn_fwrite(se: 'SymbolicExecutor', pstate: 'ProcessState'):
    """
    The fwrite behavior.
    """
    logger.debug('fwrite hooked')

    # Get arguments
    arg0 = pstate.get_argument_value(0)
    arg1 = pstate.get_argument_value(1)
    arg2 = pstate.get_argument_value(2)
    arg3 = pstate.get_argument_value(3)
    size = arg1 * arg2
    data = pstate.memory.read(arg0, size)

    if pstate.file_descriptor_exists(arg3):
        fdesc = pstate.get_file_descriptor(arg3)
        if arg3 == 0:
            return 0
        elif arg3 == 1:
            if se.config.pipe_stdout:
                fdesc.fd.buffer.write(data)
                fdesc.fd.flush()
        elif arg3 == 2:
            if se.config.pipe_stderr:
                fdesc.fd.buffer.write(data)
                fdesc.fd.flush()
        else:
            fdesc.fd.write(data)
    else:
        return 0

    # Return value
    return size


def rtn_write(se: 'SymbolicExecutor', pstate: 'ProcessState'):
    """
    The write behavior.
    """
    logger.debug('write hooked')

    # Get arguments
    fd = pstate.get_argument_value(0)
    buf = pstate.get_argument_value(1)
    size = pstate.get_argument_value(2)
    data = pstate.memory.read(buf, size)

    if pstate.file_descriptor_exists(fd):
        fdesc = pstate.get_file_descriptor(fd)
        if fd == 0:
            return 0
        elif fd == 1:
            if se.config.pipe_stdout:
                fdesc.fd.buffer.write(data)
                fdesc.fd.flush()
        elif fd == 2:
            if se.config.pipe_stderr:
                fdesc.fd.buffer.write(data)
                fdesc.fd.flush()
        else:
            fdesc.fd.write(data)
    else:
        return 0

    # Return value
    return size


def rtn_gettimeofday(se: 'SymbolicExecutor', pstate: 'ProcessState'):
    """
    The gettimeofday behavior.
    """
    logger.debug('gettimeofday hooked')

    # Get arguments
    tv = pstate.get_argument_value(0)
    tz = pstate.get_argument_value(1)

    if tv == 0:
        return pstate.minus_one

    if pstate.time_inc_coefficient:
        t = pstate.time
    else:
        t = time.time()

    s = pstate.ptr_size
    pstate.memory.write_ptr(tv, int(t))
    pstate.memory.write_ptr(tv+pstate.ptr_size, int(t * 1000000))

    # Return value
    return 0


def rtn_malloc(se: 'SymbolicExecutor', pstate: 'ProcessState'):
    """
    The malloc behavior.
    """
    logger.debug('malloc hooked')

    # Get arguments
    size = pstate.get_argument_value(0)
    ptr = pstate.heap_allocator.alloc(size)

    # Return value
    return ptr


def rtn_open(se: 'SymbolicExecutor', pstate: 'ProcessState'):
    """
    The open behavior.
    """
    logger.debug('open hooked')

    # Get arguments
    arg0 = pstate.get_argument_value(0)  # const char *pathname
    flags = pstate.get_argument_value(1)  # int flags
    mode = pstate.get_argument_value(2)  # we ignore it
    arg0s = pstate.memory.read_string(arg0)

    # Concretize the whole path name
    pstate.concretize_memory_bytes(arg0, len(arg0s)+1)  # Concretize the whole string + \0

    # We use flags as concrete value
    pstate.concretize_argument(1)

    # Use the flags to open the file in the write mode.
    mode = ""
    if (flags & 0xFF) == 0x00:   # O_RDONLY
        mode = "r"
    elif (flags & 0xFF) == 0x01: # O_WRONLY
        mode = "w"
    elif (flags & 0xFF) == 0x02: # O_RDWR
        mode = "r+"

    if flags & 0x0100: # O_CREAT
        mode += "x"
    if flags & 0x0200: # O_APPEND
        mode = "a"  # replace completely value

    # enforce using binary mode for open
    mode += "b"

    if se.seed.is_file_defined(arg0s) and "r" in mode:  # input file and opened in reading
        logger.info(f"opening an input file: {arg0s}")
        # Program is opening an input
        data = se.seed.get_file_input(arg0s)
        filedesc = pstate.create_file_descriptor(arg0s, io.BytesIO(data))
        return filedesc.id
    else:
        # Try to open it as a regular file
        try:
            fd = open(arg0s, mode)  # use the mode here
            filedesc = pstate.create_file_descriptor(arg0s, fd)
            return filedesc.id
        except Exception as e:
            logger.debug(f"Failed to open {arg0s} {e}")
            return pstate.minus_one

def rtn_realloc(se: 'SymbolicExecutor', pstate: 'ProcessState'):
    """
    The realloc behavior.
    """
    logger.debug('realloc hooked')

    # Get arguments
    oldptr = pstate.get_argument_value(0)
    size = pstate.get_argument_value(1)

    if oldptr == 0:
        # malloc behaviour
        ptr = pstate.heap_allocator.alloc(size)
        return ptr

    ptr = pstate.heap_allocator.alloc(size)
    if ptr == 0: return ptr

    if ptr not in pstate.heap_allocator.alloc_pool:
        logger.warning("Invalid ptr passed to realloc")
        pstate.heap_allocator.free(ptr) # This will raise an error

    old_memmap = pstate.heap_allocator.alloc_pool[oldptr]
    old_size = old_memmap.size
    size_to_copy = min(size, old_size)

    #data = pstate.memory.read(oldptr, size_to_copy)
    #pstate.memory.write(ptr, data)

    # Copy bytes symbolically
    for index in range(size_to_copy):
        sym_c = pstate.read_symbolic_memory_byte(oldptr+index)
        pstate.write_symbolic_memory_byte(ptr+index, sym_c)

    pstate.heap_allocator.free(oldptr)

    return ptr


def rtn_memcmp(se: 'SymbolicExecutor', pstate: 'ProcessState'):
    """
    The memcmp behavior.
    """
    logger.debug('memcmp hooked')

    s1 = pstate.get_argument_value(0)
    s2 = pstate.get_argument_value(1)
    size = pstate.get_argument_value(2)

    ptr_bit_size = pstate.ptr_bit_size

    ast = pstate.actx
    res = ast.bv(0, ptr_bit_size)

    # We constrain the logical value of size
    pstate.concretize_argument(2)

    for index in range(size):
        cells1 = pstate.read_symbolic_memory_byte(s1+index).getAst()
        cells2 = pstate.read_symbolic_memory_byte(s2+index).getAst()
        res = res + ast.ite(
                        cells1 == cells2,
                        ast.bv(0, ptr_bit_size),
                        ast.bv(1, ptr_bit_size)
                    )

    return res


def rtn_memcpy(se: 'SymbolicExecutor', pstate: 'ProcessState'):
    """
    The memcpy behavior.
    """
    logger.debug('memcpy hooked')

    dst, dst_ast = pstate.get_full_argument(0)
    src = pstate.get_argument_value(1)
    cnt = pstate.get_argument_value(2)

    # We constrain the logical value of size
    pstate.concretize_argument(2)

    for index in range(cnt):
        # Read symbolic src value and copy symbolically in dst
        sym_src = pstate.read_symbolic_memory_byte(src+index)
        pstate.write_symbolic_memory_byte(dst+index, sym_src)

    return dst_ast


def rtn_mempcpy(se: 'SymbolicExecutor', pstate: 'ProcessState'):
    """
    The mempcpy behavior.
    """
    logger.debug('mempcpy hooked')

    dst, dst_ast = pstate.get_full_argument(0)
    src = pstate.get_argument_value(1)
    cnt = pstate.get_argument_value(2)

    # We constrain the logical value of size
    pstate.concretize_argument(2)

    for index in range(cnt):
        # Read symbolic src value and copy symbolically in dst
        sym_src = pstate.read_symbolic_memory_byte(src+index)
        pstate.write_symbolic_memory_byte(dst+index, sym_src)

    return dst + cnt


def rtn_memmem(se: 'SymbolicExecutor', pstate: 'ProcessState'):
    """
    The memmem behavior.
    """
    logger.debug('memmem hooked')

    haystack    = pstate.get_argument_value(0)  # const void*
    haystacklen = pstate.get_argument_value(1)  # size_t
    needle      = pstate.get_argument_value(2)  # const void *
    needlelen   = pstate.get_argument_value(3)  # size_t

    s1 = pstate.memory.read(haystack, haystacklen)  # haystack
    s2 = pstate.memory.read(needle, needlelen)      # needle

    offset = s1.find(s2)
    if offset == -1:
        #FIXME: faut s'assurer que le marquer dans le string
        return NULL_PTR

    for i, c in enumerate(s2):
        c1 = pstate.read_symbolic_memory_byte(haystack+offset+i)
        c2 = pstate.read_symbolic_memory_byte(needle+i)
        pstate.push_constraint(c1.getAst() == c2.getAst())

    # FIXME: à reflechir si on doit contraindre offset ou pas

    # faut s'assurer que le marquer est bien présent à l'offset trouvé
    return haystack + offset


def rtn_memmove(se: 'SymbolicExecutor', pstate: 'ProcessState'):
    """
    The memmove behavior.
    """
    logger.debug('memmove hooked')

    dst, dst_ast = pstate.get_full_argument(0)
    src = pstate.get_argument_value(1)
    cnt = pstate.get_argument_value(2)

    # We constrain the logical value of cnt
    pstate.concretize_argument(2)

    # TODO: What if cnt is symbolic ?
    for index in range(cnt):
        # Read symbolic src value and copy symbolically in dst
        sym_src = pstate.read_symbolic_memory_byte(src+index)
        pstate.write_symbolic_memory_byte(dst+index, sym_src)

    return dst_ast


def rtn_memset(se: 'SymbolicExecutor', pstate: 'ProcessState'):
    """
    The memset behavior.
    """
    logger.debug('memset hooked')

    dst, dst_ast = pstate.get_full_argument(0)
    src, src_ast = pstate.get_full_argument(1)
    size = pstate.get_argument_value(2)

    # We constrain the logical value of size
    pstate.concretize_argument(2)

    sym_cell = pstate.actx.extract(7, 0, src_ast.getAst())

    # TODO: What if size is symbolic ?
    for index in range(size):
        pstate.write_symbolic_memory_byte(dst+index, sym_cell)

    return dst_ast


def rtn_printf(se: 'SymbolicExecutor', pstate: 'ProcessState'):
    """
    The printf behavior.
    """
    logger.debug('printf hooked')

    # Get arguments
    arg0 = pstate.get_argument_value(0)

    arg0f = pstate.get_format_string(arg0)
    nbArgs = arg0f.count("{")
    args = pstate.get_format_arguments(arg0, [pstate.get_argument_value(x) for x in range(1, nbArgs+1)])
    try:
        s = arg0f.format(*args)
    except:
        # FIXME: Les chars UTF8 peuvent foutre le bordel. Voir avec ground-truth/07.input
        logger.warning('Something wrong, probably UTF-8 string')
        s = ""

    if se.config.pipe_stdout:
        stdout = pstate.get_file_descriptor(1)
        stdout.fd.write(s)
        stdout.fd.flush()

    # Return value
    return len(s)


def rtn_pthread_create(se: 'SymbolicExecutor', pstate: 'ProcessState'):
    """
    The pthread_create behavior.
    """
    logger.debug('pthread_create hooked')

    # Get arguments
    arg0 = pstate.get_argument_value(0) # pthread_t *thread
    arg1 = pstate.get_argument_value(1) # const pthread_attr_t *attr
    arg2 = pstate.get_argument_value(2) # void *(*start_routine) (void *)
    arg3 = pstate.get_argument_value(3) # void *arg

    th = pstate.spawn_new_thread(arg2, arg3)

    # Save out the thread id
    pstate.memory.write_ptr(arg0, th.tid)

    # Return value
    return 0


def rtn_pthread_exit(se: 'SymbolicExecutor', pstate: 'ProcessState'):
    """
    The pthread_exit behavior.
    """
    logger.debug('pthread_exit hooked')

    # Get arguments
    arg0 = pstate.get_argument_value(0)

    # Kill the thread
    pstate.current_thread.kill()

    # FIXME: I guess the thread exiting never return, so should not continue
    # FIXME: iterating instructions

    # Return value
    return None


def rtn_pthread_join(se: 'SymbolicExecutor', pstate: 'ProcessState'):
    """
    The pthread_join behavior.
    """
    logger.debug('pthread_join hooked')

    # Get arguments
    arg0 = pstate.get_argument_value(0)
    arg1 = pstate.get_argument_value(1)

    if arg0 in pstate.threads:
        pstate.current_thread.join_thread(arg0)
        logger.info(f"Thread id {pstate.current_thread.tid} joined thread id {arg0}")
    else:
        pstate.current_thread.cancel_join()
        logger.debug(f"Thread id {arg0} already destroyed")

    # Return value
    return 0


def rtn_pthread_mutex_destroy(se: 'SymbolicExecutor', pstate: 'ProcessState'):
    """
    The pthread_mutex_destroy behavior.
    """
    logger.debug('pthread_mutex_destroy hooked')

    # Get arguments
    arg0 = pstate.get_argument_value(0)  # pthread_mutex_t *restrict mutex
    pstate.memory.write_ptr(arg0, pstate.PTHREAD_MUTEX_INIT_MAGIC)

    return 0


def rtn_pthread_mutex_init(se: 'SymbolicExecutor', pstate: 'ProcessState'):
    """
    The pthread_mutex_init behavior.
    """
    logger.debug('pthread_mutex_init hooked')

    # Get arguments
    arg0 = pstate.get_argument_value(0)  # pthread_mutex_t *restrict mutex
    arg1 = pstate.get_argument_value(1)  # const pthread_mutexattr_t *restrict attr)

    pstate.memory.write_ptr(arg0, pstate.PTHREAD_MUTEX_INIT_MAGIC)

    return 0


def rtn_pthread_mutex_lock(se: 'SymbolicExecutor', pstate: 'ProcessState'):
    """
    The pthread_mutex_lock behavior.
    """
    logger.debug('pthread_mutex_lock hooked')

    # Get arguments
    arg0 = pstate.get_argument_value(0)  # pthread_mutex_t *mutex
    mutex = pstate.memory.read_ptr(arg0)  # deref pointer and read a uint64 int

    # If the thread has been initialized and unused, define the tid has lock
    if mutex == pstate.PTHREAD_MUTEX_INIT_MAGIC:
        logger.debug('mutex unlocked')
        pstate.memory.write_ptr(arg0, pstate.current_thread.tid)

    # The mutex is locked and we are not allowed to continue the execution
    elif mutex != pstate.current_thread.tid:
        logger.debug('mutex locked')
        pstate.mutex_locked = True

    # Return value
    return 0


def rtn_pthread_mutex_unlock(se: 'SymbolicExecutor', pstate: 'ProcessState'):
    """
    The pthread_mutex_unlock behavior.
    """
    logger.debug('pthread_mutex_unlock hooked')

    # Get arguments
    arg0 = pstate.get_argument_value(0)  # pthread_mutex_t *mutex

    pstate.memory.write_ptr(arg0, pstate.PTHREAD_MUTEX_INIT_MAGIC)

    # Return value
    return 0


def rtn_puts(se: 'SymbolicExecutor', pstate: 'ProcessState'):
    """
    The puts behavior.
    """
    logger.debug('puts hooked')

    arg0 = pstate.get_string_argument(0)

    # Get arguments
    if se.config.pipe_stdout:  # Only perform printing if pipe_stdout activated
        sys.stdout.write(arg0 + '\n')
        sys.stdout.flush()

    # Return value
    return len(arg0) + 1


def rtn_rand(se: 'SymbolicExecutor', pstate: 'ProcessState'):
    """
    The rand behavior.
    """
    logger.debug('rand hooked')
    return random.randrange(0, 0xffffffff)


def rtn_read(se: 'SymbolicExecutor', pstate: 'ProcessState'):
    """
    The read behavior.
    """
    logger.debug('read hooked')

    # Get arguments
    fd   = pstate.get_argument_value(0)
    buff = pstate.get_argument_value(1)
    size, size_ast = pstate.get_full_argument(2)

    if size_ast.isSymbolized():
        logger.warning(f'Reading from the file descriptor ({fd}) with a symbolic size')

    pstate.concretize_argument(0)

    if pstate.file_descriptor_exists(fd):
        filedesc = pstate.get_file_descriptor(fd)
        offset = filedesc.offset
        data = filedesc.read(size)

        if filedesc.is_input_fd(): # Reading into input
            # if we started from empty seed simulate reading `size` amount of data
            if se.seed.is_raw() and se.seed.is_bootstrap_seed() and not data:
                data = b'\x00' * size

            if len(data) == size:  # if `size` limited the fgets its an indirect constraint
                pstate.push_constraint(size_ast.getAst() == size)

            se.inject_symbolic_file_memory(buff, filedesc.name, data, offset)
            logger.debug(f"read in (input) {filedesc.name} = {repr(data)}")
        else:
            pstate.concretize_argument(2)
            pstate.memory.write(buff, data)

        return len(data)
    else:
        logger.warning(f'File descriptor ({fd}) not found')
        return pstate.minus_one  # todo: set errno


def rtn_getchar(se: 'SymbolicExecutor', pstate: 'ProcessState'):
    """
    The getchar behavior.
    """
    logger.debug('getchar hooked')

    # Get arguments
    filedesc = pstate.get_file_descriptor(0) # stdin
    offset = filedesc.offset

    data = filedesc.read(1)
    if data:
        if filedesc.is_input_fd():  # Reading into input
            data = ord(data) # convert to integer
            se.inject_symbolic_file_register(pstate.return_register, filedesc.name, data, offset)
            logger.debug(f"read in {filedesc.name} = {repr(data)}")
        return data
    else:
        return pstate.minus_one


def rtn_sem_destroy(se: 'SymbolicExecutor', pstate: 'ProcessState'):
    """
    The sem_destroy behavior.
    """
    logger.debug('sem_destroy hooked')

    # Get arguments
    arg0 = pstate.get_argument_value(0)  # sem_t *sem

    # Destroy the semaphore with the value
    pstate.memory.write_ptr(arg0, 0)

    # Return success
    return 0


def rtn_sem_getvalue(se: 'SymbolicExecutor', pstate: 'ProcessState'):
    """
    The sem_getvalue behavior.
    """
    logger.debug('sem_getvalue hooked')

    # Get arguments
    arg0 = pstate.get_argument_value(0)  # sem_t *sem
    arg1 = pstate.get_argument_value(1)  # int *sval

    value = pstate.memory.read_ptr(arg0)  # deref pointer

    # Set the semaphore's value into the output
    pstate.memory.write_dword(arg1, value)  # WARNING: read uint64 to uint32

    # Return success
    return 0


def rtn_sem_init(se: 'SymbolicExecutor', pstate: 'ProcessState'):
    """
    The sem_init behavior.
    """
    logger.debug('sem_init hooked')

    # Get arguments
    arg0 = pstate.get_argument_value(0)  # sem_t *sem
    arg1 = pstate.get_argument_value(1)  # int pshared
    arg2 = pstate.get_argument_value(2)  # unsigned int value

    # Init the semaphore with the value
    pstate.memory.write_ptr(arg0, arg2)

    # Return success
    return 0


def rtn_sem_post(se: 'SymbolicExecutor', pstate: 'ProcessState'):
    """
    The sem_post behavior.
    """
    logger.debug('sem_post hooked')

    arg0 = pstate.get_argument_value(0)  # sem_t *sem

    # increments (unlocks) the semaphore pointed to by sem
    value = pstate.memory.read_ptr(arg0)
    pstate.memory.write_ptr(arg0, value + 1)

    # Return success
    return 0


def rtn_sem_timedwait(se: 'SymbolicExecutor', pstate: 'ProcessState'):
    """
    The sem_timedwait behavior.
    """
    logger.debug('sem_timedwait hooked')

    arg0 = pstate.get_argument_value(0)  # sem_t *sem
    arg1 = pstate.get_argument_value(1)  # const struct timespec *abs_timeout

    # sem_timedwait() is the same as sem_wait(), except that abs_timeout specifies a limit
    # on the amount of time that the call should block if the decrement cannot be immediately
    # performed. The abs_timeout argument points to a structure that specifies an absolute
    # timeout in seconds and nanoseconds since the Epoch, 1970-01-01 00:00:00 +0000 (UTC).
    # This structure is defined as follows:
    #
    #     struct timespec {
    #         time_t tv_sec;      /* Seconds */
    #         long   tv_nsec;     /* Nanoseconds [0 .. 999999999] */
    #     };
    #
    # If the timeout has already expired by the time of the call, and the semaphore could not be
    # locked immediately, then sem_timedwait() fails with a timeout error (errno set to ETIMEDOUT).
    #
    # If  the operation can be performed immediately, then sem_timedwait() never fails with a
    # timeout error, regardless of the value of abs_timeout.  Furthermore, the validity of
    # abs_timeout is not checked in this case.

    # TODO: Take into account the abs_timeout argument
    value = pstate.memory.read_ptr(arg0)
    if value > 0:
        logger.debug('semaphore still not locked')
        pstate.memory.write_ptr(arg0, value - 1)
        pstate.semaphore_locked = False
    else:
        logger.debug('semaphore locked')
        pstate.semaphore_locked = True

    # Return success
    return 0


def rtn_sem_trywait(se: 'SymbolicExecutor', pstate: 'ProcessState'):
    """
    The sem_trywait behavior.
    """
    logger.debug('sem_trywait hooked')

    arg0 = pstate.get_argument_value(0)  # sem_t *sem

    # sem_trywait()  is  the  same as sem_wait(), except that if the decrement
    # cannot be immediately performed, then call returns an error (errno set to
    # EAGAIN) instead of blocking.
    value = pstate.memory.read_ptr(arg0)
    if value > 0:
        logger.debug('semaphore still not locked')
        pstate.memory.write_ptr(arg0, value - 1)
        pstate.semaphore_locked = False
    else:
        logger.debug('semaphore locked but continue')
        pstate.semaphore_locked = False

        # Setting errno to EAGAIN (3406)
        segs = pstate.memory.find_map(pstate.EXTERN_SEG)
        if segs:
            map = segs[0]
            ERRNO = map.start + map.size - 4  # Point is last int of the mapping
            pstate.memory.write_dword(ERRNO, 3406)
        else:
            assert False

        # Return -1
        return pstate.minus_one

    # Return success
    return 0


def rtn_sem_wait(se: 'SymbolicExecutor', pstate: 'ProcessState'):
    """
    The sem_wait behavior.
    """
    logger.debug('sem_wait hooked')

    arg0 = pstate.get_argument_value(0)  # sem_t *sem

    # decrements (locks) the semaphore pointed to by sem. If the semaphore's value
    # is greater than zero, then the decrement proceeds, and the function returns,
    # immediately. If the semaphore currently has the value zero, then the call blocks
    # until either it becomes possible to perform the decrement (i.e., the semaphore
    # value rises above zero).
    value = pstate.memory.read_ptr(arg0)
    if value > 0:
        logger.debug('semaphore still not locked')
        pstate.memory.write_ptr(arg0, value - 1)
        pstate.semaphore_locked = False
    else:
        logger.debug('semaphore locked')
        pstate.semaphore_locked = True

    # Return success
    return 0


def rtn_sleep(se: 'SymbolicExecutor', pstate: 'ProcessState'):
    """
    The sleep behavior.
    """
    logger.debug('sleep hooked')

    # Get arguments
    if not se.config.skip_sleep_routine:
        t = pstate.get_argument_value(0)
        time.sleep(t)

    # Return value
    return 0


def rtn_sprintf(se: 'SymbolicExecutor', pstate: 'ProcessState'):
    """
    The sprintf behavior.
    """
    logger.debug('sprintf hooked')

    # Get arguments
    buff = pstate.get_argument_value(0)
    arg0 = pstate.get_argument_value(1)

    try:
        arg0f = pstate.get_format_string(arg0)
        nbArgs = arg0f.count("{")
        args = pstate.get_format_arguments(arg0, [pstate.get_argument_value(x) for x in range(2, nbArgs+2)])
        s = arg0f.format(*args)
    except:
        # FIXME: Les chars UTF8 peuvent foutre le bordel. Voir avec ground-truth/07.input
        logger.warning('Something wrong, probably UTF-8 string')
        s = ""

    # FIXME: todo

    # FIXME: THIS SEEMS NOT OK
    # for index, c in enumerate(s):
    #     pstate.tt_ctx.concretizeMemory(buff + index)
    #     pstate.tt_ctx.setConcreteMemoryValue(buff + index, ord(c))
    #     pstate.tt_ctx.pushPathConstraint(pstate.tt_ctx.getMemoryAst(MemoryAccess(buff + index, 1)) == ord(c))
    #
    # # including the terminating null byte ('\0')
    # pstate.tt_ctx.setConcreteMemoryValue(buff + len(s), 0x00)
    # pstate.tt_ctx.pushPathConstraint(pstate.tt_ctx.getMemoryAst(MemoryAccess(buff + len(s), 1)) == 0x00)

    return len(s)


def rtn_strcasecmp(se: 'SymbolicExecutor', pstate: 'ProcessState'):
    """
    The strcasecmp behavior.
    """
    logger.debug('strcasecmp hooked')

    s1 = pstate.get_argument_value(0)
    s2 = pstate.get_argument_value(1)
    size = min(len(pstate.memory.read_string(s1)), len(pstate.memory.read_string(s2)) + 1)

    #s = s1 if len(pstate.memory.read_string(s1)) < len(pstate.memory.read_string(s2)) else s2
    #for i in range(size):
    #    pstate.tt_ctx.pushPathConstraint(pstate.tt_ctx.getMemoryAst(MemoryAccess(s1 + i, CPUSIZE.BYTE)) != 0x00)
    #    pstate.tt_ctx.pushPathConstraint(pstate.tt_ctx.getMemoryAst(MemoryAccess(s2 + i, CPUSIZE.BYTE)) != 0x00)
    #pstate.tt_ctx.pushPathConstraint(pstate.tt_ctx.getMemoryAst(MemoryAccess(s + size, CPUSIZE.BYTE)) == 0x00)
    #pstate.tt_ctx.pushPathConstraint(pstate.tt_ctx.getMemoryAst(MemoryAccess(s1 + len(pstate.memory.read_string(s1)), CPUSIZE.BYTE)) == 0x00)
    #pstate.tt_ctx.pushPathConstraint(pstate.tt_ctx.getMemoryAst(MemoryAccess(s2 + len(pstate.memory.read_string(s2)), CPUSIZE.BYTE)) == 0x00)

    # FIXME: Il y a des truc chelou avec le +1 et le logic ci-dessous

    ptr_bit_size = pstate.ptr_bit_size
    ast = pstate.actx
    res = ast.bv(0, pstate.ptr_bit_size)
    for index in range(size):
        cells1 = pstate.read_symbolic_memory_byte(s1 + index).getAst()
        cells2 = pstate.read_symbolic_memory_byte(s2 + index).getAst()
        cells1 = ast.ite(ast.land([cells1 >= ord('a'), cells1 <= ord('z')]), cells1 - 32, cells1) # upper case
        cells2 = ast.ite(ast.land([cells2 >= ord('a'), cells2 <= ord('z')]), cells2 - 32, cells2) # upper case
        res = res + ast.ite(cells1 == cells2, ast.bv(0, ptr_bit_size), ast.bv(1, ptr_bit_size))

    return res


def rtn_strchr(se: 'SymbolicExecutor', pstate: 'ProcessState'):
    """
    The strchr behavior.
    """
    logger.debug('strchr hooked')

    string = pstate.get_argument_value(0)
    char   = pstate.get_argument_value(1)
    ast    = pstate.actx
    ptr_bit_size = pstate.ptr_bit_size

    def rec(res, deep, maxdeep):
        if deep == maxdeep:
            return res
        cell = pstate.read_symbolic_memory_byte(string + deep).getAst()
        res  = ast.ite(cell == (char & 0xff), ast.bv(string + deep, ptr_bit_size), rec(res, deep + 1, maxdeep))
        return res

    sze = len(pstate.memory.read_string(string))
    res = rec(ast.bv(0, ptr_bit_size), 0, sze)

    for i, c in enumerate(pstate.memory.read_string(string)):
        pstate.push_constraint(pstate.read_symbolic_memory_byte(string+i).getAst() != 0x00)
    pstate.push_constraint(pstate.read_symbolic_memory_byte(string+sze).getAst() == 0x00)

    return res


def rtn_strcmp(se: 'SymbolicExecutor', pstate: 'ProcessState'):
    """
    The strcmp behavior.
    """
    logger.debug('strcmp hooked')

    s1 = pstate.get_argument_value(0)
    s2 = pstate.get_argument_value(1)
    size = min(len(pstate.memory.read_string(s1)), len(pstate.memory.read_string(s2))) + 1
    
    #s = s1 if len(pstate.memory.read_string(s1)) <= len(pstate.memory.read_string(s2)) else s2
    #for i in range(size):
    #    pstate.tt_ctx.pushPathConstraint(pstate.tt_ctx.getMemoryAst(MemoryAccess(s1 + i, CPUSIZE.BYTE)) != 0x00)
    #    pstate.tt_ctx.pushPathConstraint(pstate.tt_ctx.getMemoryAst(MemoryAccess(s2 + i, CPUSIZE.BYTE)) != 0x00)
    #pstate.tt_ctx.pushPathConstraint(pstate.tt_ctx.getMemoryAst(MemoryAccess(s + size, CPUSIZE.BYTE)) == 0x00)
    #pstate.tt_ctx.pushPathConstraint(pstate.tt_ctx.getMemoryAst(MemoryAccess(s1 + len(pstate.memory.read_string(s1)), CPUSIZE.BYTE)) == 0x00)
    #pstate.tt_ctx.pushPathConstraint(pstate.tt_ctx.getMemoryAst(MemoryAccess(s2 + len(pstate.memory.read_string(s2)), CPUSIZE.BYTE)) == 0x00)

    # FIXME: Il y a des truc chelou avec le +1 et le logic ci-dessous

    ptr_bit_size = pstate.ptr_bit_size
    ast = pstate.actx
    res = ast.bv(0, ptr_bit_size)
    for index in range(size):
        cells1 = pstate.read_symbolic_memory_byte(s1 + index).getAst()
        cells2 = pstate.read_symbolic_memory_byte(s2 + index).getAst()
        res = res + ast.ite(cells1 == cells2, ast.bv(0, ptr_bit_size), ast.bv(1, ptr_bit_size))

    return res


def rtn_strcpy(se: 'SymbolicExecutor', pstate: 'ProcessState'):
    """
    The strcpy behavior.
    """
    logger.debug('strcpy hooked')

    dst  = pstate.get_argument_value(0)
    src  = pstate.get_argument_value(1)
    src_str = pstate.memory.read_string(src)
    size = len(src_str)

    # constraint src buff to be != \00 and last one to be \00 (indirectly concretize length)
    for i, c in enumerate(src_str):
        pstate.push_constraint(pstate.read_symbolic_memory_byte(src + i).getAst() != 0x00)
    pstate.push_constraint(pstate.read_symbolic_memory_byte(src + size).getAst() == 0x00)

    # Copy symbolically bytes (including \00)
    for index in range(size+1):
        sym_c = pstate.read_symbolic_memory_byte(src+index)
        pstate.write_symbolic_memory_byte(dst+index, sym_c)

    return dst


def rtn_strdup(se: 'SymbolicExecutor', pstate: 'ProcessState'):
    """
    The strdup behavior.
    """
    logger.debug('strdup hooked')

    s  = pstate.get_argument_value(0)
    s_str = pstate.memory.read_string(s)
    size = len(s_str)

    #print(f"strdup s={s:#x} s_str={s_str} size={size}")

    # constrain src buff to be != \00 and last one to be \00 (indirectly concretize length)
    for i, c in enumerate(s_str):
        pstate.push_constraint(pstate.read_symbolic_memory_byte(s + i).getAst() != 0x00)
    pstate.push_constraint(pstate.read_symbolic_memory_byte(s + size).getAst() == 0x00)


    # Malloc a chunk
    ptr = pstate.heap_allocator.alloc(size + 1)

    # Copy symbolically bytes (including \00)
    for index in range(size+1):
        sym_c = pstate.read_symbolic_memory_byte(s+index)
        pstate.write_symbolic_memory_byte(ptr+index, sym_c)

    return ptr


def rtn_strerror(se: 'SymbolicExecutor', pstate: 'ProcessState'):
    """
    The strerror behavior.

    :param se: The current symbolic execution instance
    :param pstate: The current process state
    :return: a concrete value
    """
    logger.debug('strerror hooked')

    sys_errlist = [
        b"Success",
        b"Operation not permitted",
        b"No such file or directory",
        b"No such process",
        b"Interrupted system call",
        b"Input/output error",
        b"No such device or address",
        b"Argument list too long",
        b"Exec format error",
        b"Bad file descriptor",
        b"No child processes",
        b"Resource temporarily unavailable",
        b"Cannot allocate memory",
        b"Permission denied",
        b"Bad address",
        b"Block device required",
        b"Device or resource busy",
        b"File exists",
        b"Invalid cross-device link",
        b"No such device",
        b"Not a directory",
        b"Is a directory",
        b"Invalid argument",
        b"Too many open files in system",
        b"Too many open files",
        b"Inappropriate ioctl for device",
        b"Text file busy",
        b"File too large",
        b"No space left on device",
        b"Illegal seek",
        b"Read-only file system",
        b"Too many links",
        b"Broken pipe",
        b"Numerical argument out of domain",
        b"Numerical result out of range",
        b"Resource deadlock avoided",
        b"File name too long",
        b"No locks available",
        b"Function not implemented",
        b"Directory not empty",
        b"Too many levels of symbolic links",
        None,
        b"No message of desired type",
        b"Identifier removed",
        b"Channel number out of range",
        b"Level 2 not synchronized",
        b"Level 3 halted",
        b"Level 3 reset",
        b"Link number out of range",
        b"Protocol driver not attached",
        b"No CSI structure available",
        b"Level 2 halted",
        b"Invalid exchange",
        b"Invalid request descriptor",
        b"Exchange full",
        b"No anode",
        b"Invalid request code",
        b"Invalid slot",
        None,
        b"Bad font file format",
        b"Device not a stream",
        b"No data available",
        b"Timer expired",
        b"Out of streams resources",
        b"Machine is not on the network",
        b"Package not installed",
        b"Object is remote",
        b"Link has been severed",
        b"Advertise error",
        b"Srmount error",
        b"Communication error on send",
        b"Protocol error",
        b"Multihop attempted",
        b"RFS specific error",
        b"Bad message",
        b"Value too large for defined data type",
        b"Name not unique on network",
        b"File descriptor in bad state",
        b"Remote address changed",
        b"Can not access a needed shared library",
        b"Accessing a corrupted shared library",
        b".lib section in a.out corrupted",
        b"Attempting to link in too many shared libraries",
        b"Cannot exec a shared library directly",
        b"Invalid or incomplete multibyte or wide character",
        b"Interrupted system call should be restarted",
        b"Streams pipe error",
        b"Too many users",
        b"Socket operation on non-socket",
        b"Destination address required",
        b"Message too long",
        b"Protocol wrong type for socket",
        b"Protocol not available",
        b"Protocol not supported",
        b"Socket type not supported",
        b"Operation not supported",
        b"Protocol family not supported",
        b"Address family not supported by protocol",
        b"Address already in use",
        b"Cannot assign requested address",
        b"Network is down",
        b"Network is unreachable",
        b"Network dropped connection on reset",
        b"Software caused connection abort",
        b"Connection reset by peer",
        b"No buffer space available",
        b"Transport endpoint is already connected",
        b"Transport endpoint is not connected",
        b"Cannot send after transport endpoint shutdown",
        b"Too many references: cannot splice",
        b"Connection timed out",
        b"Connection refused",
        b"Host is down",
        b"No route to host",
        b"Operation already in progress",
        b"Operation now in progress",
        b"Stale NFS file handle",
        b"Structure needs cleaning",
        b"Not a XENIX named type file",
        b"No XENIX semaphores available",
        b"Is a named type file",
        b"Remote I/O error",
        b"Disk quota exceeded",
        b"No medium found",
        b"Wrong medium type",
        b"Operation canceled",
        b"Required key not available",
        b"Key has expired",
        b"Key has been revoked",
        b"Key was rejected by service",
        b"Owner died",
        b"State not recoverable"
    ]

    # Get arguments
    errnum = pstate.get_argument_value(0)
    try:
        str = sys_errlist[errnum]
    except:
        # invalid errnum
        str = b'Error'

    # TODO: We allocate the string at every hit of this function with a
    # potential memory leak. We should allocate the sys_errlist only once
    # and then refer to this table instead of allocate string.
    ptr = pstate.heap_allocator.alloc(len(str) + 1)
    pstate.memory.write(ptr, str + b'\0')

    return ptr


def rtn_strlen(se: 'SymbolicExecutor', pstate: 'ProcessState'):
    """
    The strlen behavior.
    """
    logger.debug('strlen hooked')

    ptr_bit_size = pstate.ptr_bit_size

    # Get arguments
    s = pstate.get_argument_value(0)
    ast = pstate.actx

    # FIXME: Not so sure its is really the strlen semantic
    def rec(res, s, deep, maxdeep):
        if deep == maxdeep:
            return res
        cell = pstate.read_symbolic_memory_byte(s+deep).getAst()
        res  = ast.ite(cell == 0x00, ast.bv(deep, ptr_bit_size), rec(res, s, deep + 1, maxdeep))
        return res

    sze = len(pstate.memory.read_string(s))
    res = ast.bv(sze, ptr_bit_size)
    res = rec(res, s, 0, sze)

    # FIXME: That routine should do something like below to be SOUND !
    # for i, c in enumerate(pstate.memory.read_string(src)):
    #     pstate.push_constraint(pstate.read_symbolic_memory_byte(src + i) != 0x00)
    # pstate.push_constraint(pstate.read_symbolic_memory_byte(src + size) == 0x00)

    pstate.push_constraint(pstate.read_symbolic_memory_byte(s+sze).getAst() == 0x00)

    return res


def rtn_strncasecmp(se: 'SymbolicExecutor', pstate: 'ProcessState'):
    """
    The strncasecmp behavior.
    """
    logger.debug('strncasecmp hooked')

    s1 = pstate.get_argument_value(0)
    s2 = pstate.get_argument_value(1)
    sz = pstate.get_argument_value(2)
    maxlen = min(sz, min(len(pstate.memory.read_string(s1)), len(pstate.memory.read_string(s2))) + 1)

    ptr_bit_size = pstate.ptr_bit_size

    ast = pstate.actx
    res = ast.bv(0, ptr_bit_size)
    for index in range(maxlen):
        cells1 = pstate.read_symbolic_memory_byte(s1 + index).getAst()
        cells2 = pstate.read_symbolic_memory_byte(s2 + index).getAst()
        cells1 = ast.ite(ast.land([cells1 >= ord('a'), cells1 <= ord('z')]), cells1 - 32, cells1) # upper case
        cells2 = ast.ite(ast.land([cells2 >= ord('a'), cells2 <= ord('z')]), cells2 - 32, cells2) # upper case
        res = res + ast.ite(cells1 == cells2, ast.bv(0, ptr_bit_size), ast.bv(1, ptr_bit_size))

    return res


def rtn_strncmp(se: 'SymbolicExecutor', pstate: 'ProcessState'):
    """
    The strncmp behavior.
    """
    logger.debug('strncmp hooked')

    s1 = pstate.get_argument_value(0)
    s2 = pstate.get_argument_value(1)
    sz = pstate.get_argument_value(2)
    maxlen = min(sz, min(len(pstate.memory.read_string(s1)), len(pstate.memory.read_string(s2))) + 1)

    ptr_bit_size = pstate.ptr_bit_size

    ast = pstate.actx
    res = ast.bv(0, ptr_bit_size)
    for index in range(maxlen):
        cells1 = pstate.read_symbolic_memory_byte(s1 + index).getAst()
        cells2 = pstate.read_symbolic_memory_byte(s2 + index).getAst()
        res = res + ast.ite(cells1 == cells2, ast.bv(0, ptr_bit_size), ast.bv(1, ptr_bit_size))

    return res


def rtn_strncpy(se: 'SymbolicExecutor', pstate: 'ProcessState'):
    """
    The strncpy behavior.
    """
    logger.debug('strncpy hooked')

    dst = pstate.get_argument_value(0)
    src = pstate.get_argument_value(1)
    cnt = pstate.get_argument_value(2)

    pstate.concretize_argument(2)

    for index in range(cnt):
        src_sym = pstate.read_symbolic_memory_byte(src+index)
        pstate.write_symbolic_memory_byte(dst+index, src_sym)

        if src_sym.getAst().evaluate() == 0:
            pstate.push_constraint(src_sym.getAst() == 0x00)
            break
        else:
            pstate.push_constraint(src_sym.getAst() != 0x00)

    return dst


def rtn_strtok_r(se: 'SymbolicExecutor', pstate: 'ProcessState'):
    """
    The strtok_r behavior.
    """
    logger.debug('strtok_r hooked')

    string  = pstate.get_argument_value(0)
    delim   = pstate.get_argument_value(1)
    saveptr = pstate.get_argument_value(2)
    saveMem = pstate.memory.read_ptr(saveptr)

    if string == 0:
        string = saveMem

    d = pstate.memory.read_string(delim)
    s = pstate.memory.read_string(string)

    tokens = re.split('[' + re.escape(d) + ']', s)

    # TODO: Make it symbolic
    for token in tokens:
        if token:
            offset = s.find(token)
            # Init the \0 at the delimiter position
            node = pstate.read_symbolic_memory_byte(string + offset + len(token)).getAst()
            try:
                pstate.push_constraint(pstate.actx.lor([node == ord(c) for c in d]))
            except:  # dafuck is that?
                pstate.push_constraint(node == ord(d))

            # Token must not contain delimiters
            for index, char in enumerate(token):
                node = pstate.read_symbolic_memory_byte(string + offset + index).getAst()
                for delim in d:
                    pstate.push_constraint(node != ord(delim))

            pstate.memory.write_char(string + offset + len(token), 0)
            # Save the pointer
            pstate.memory.write_ptr(saveptr, string + offset + len(token) + 1)
            # Return the token
            return string + offset

    return NULL_PTR


def rtn_strtoul(se: 'SymbolicExecutor', pstate: 'ProcessState'):
    """
    The strtoul behavior.
    """
    logger.debug('strtoul hooked')

    nptr   = pstate.get_argument_value(0)
    nptrs  = pstate.get_string_argument(0)
    endptr = pstate.get_argument_value(1)
    base   = pstate.get_argument_value(2)

    for i, c in enumerate(nptrs):
        pstate.push_constraint(pstate.read_symbolic_memory_byte(nptr+i).getAst() == ord(c))

    pstate.concretize_argument(2)  # Concretize base

    try:
        return int(nptrs, base)
    except:
        return 0xffffffff


def rtn_getenv(se: 'SymbolicExecutor', pstate: 'ProcessState'):
    """
    The getenv behavior.
    """
    # TODO
    name = pstate.get_argument_value(0)

    if name == 0:
        return NULL_PTR

    environ_name = pstate.memory.read_string(name)
    logger.warning(f"Target called getenv({environ_name})")
    host_env_val = os.getenv(environ_name)
    return host_env_val if host_env_val is not None else 0


#def rtn_tolower(se: 'SymbolicExecutor', pstate: 'ProcessState'):
#    # TODO
#    """
#    The tolower behavior.
#    """
#    ptr_bit_size = pstate.ptr_bit_size
#    ast = pstate.actx
#    arg_sym = pstate.get_argument_symbolic(0)
#    return rdi_sym.getAst() - 0x20

def rtn_isspace(se: 'SymbolicExecutor', pstate: 'ProcessState'):
    """
    The isspace behavior.
    """
    ptr_bit_size = pstate.ptr_bit_size
    ast = pstate.actx
    arg_sym = pstate.get_argument_symbolic(0)

    exp = arg_sym.getAst() == 0x20
    exp = ast.lor([exp, arg_sym.getAst() == 0xa])
    exp = ast.lor([exp, arg_sym.getAst() == 0x9])
    exp = ast.lor([exp, arg_sym.getAst() == 0xc])
    exp = ast.lor([exp, arg_sym.getAst() == 0xd])
    res =  ast.ite(exp, ast.bv(0, ptr_bit_size), ast.bv(1, ptr_bit_size))
    return res


def rtn_assert_fail(se: 'SymbolicExecutor', pstate: 'ProcessState'):
    """
    The __assert_fail behavior.
    """
    msg = pstate.get_argument_value(0)

    msg = pstate.memory.read_string(msg)
    logger.warning(f"__assert_fail called : {msg}")

    # Write 1 as return value of the program
    pstate.write_register(pstate.return_register, 1)
    se.abort()

def rtn_setlocale(se: 'SymbolicExecutor', pstate: 'ProcessState'):
    """
    The setlocale behavior.
    """
    logger.debug('setlocale hooked')

    category = pstate.get_argument_value(0)
    locale   = pstate.get_argument_value(1)

    if locale != 0:
        logger.warning(f"Attempt to modify Locale. Currently not supported.")
        return 0

    # This is a bit hacky but we just store the LOCALEs in the [extern] segment
    segs = pstate.memory.find_map(pstate.EXTERN_SEG)
    if segs:
        map = segs[0]
        LC_ALL = map.start + map.size - 0x20 # Point to the end of seg. But keep in mind LC_ALL is at end - 4.
    else:
        assert False
    print(f"selocale writing at {LC_ALL:#x}")

    if category == 0:
        pstate.memory.write(LC_ALL, b"en_US.UTF-8\x00")
    else:
        logger.warning(f"setlocale called with unsupported category={category}.")

    return LC_ALL


def rtn__setjmp(se: 'SymbolicExecutor', pstate: 'ProcessState'):
    """
    The _setjmp behavior. 
    """
    # TODO
    logger.warning("hooked _setjmp")
    return 0


def rtn_longjmp(se: 'SymbolicExecutor', pstate: 'ProcessState'):
    """
    The longjmp behavior.
    """
    # NOTE All the programs tested so far used `longjmp` as an error handling mechanism, right
    # before exiting. This is why, `longjmp` is currently considered an exit condition. 
    # TODO Real implementation
    logger.debug('longjmp hooked')
    pstate.stop = True

def rtn_atexit(se: 'SymbolicExecutor', pstate: 'ProcessState'):
    return 0


SUPPORTED_ROUTINES = {
    # TODO:
    #   - tolower
    #   - toupper
    '__assert_fail':           rtn_assert_fail,
    '__ctype_b_loc':           rtn_ctype_b_loc,
    '__ctype_toupper_loc':     rtn_ctype_toupper_loc,
    '__errno_location':        rtn_errno_location,
    '__fprintf_chk':           rtn___fprintf_chk,
    '__libc_start_main':       rtn_libc_start_main,
    '__stack_chk_fail':        rtn_stack_chk_fail,
    '__xstat':                 rtn_xstat,
    'abort':                   rtn_abort,
    "atexit":                  rtn_atexit,
    "__cxa_atexit":            rtn_atexit,
    'atoi':                    rtn_atoi,
    'calloc':                  rtn_calloc,
    'clock_gettime':           rtn_clock_gettime,
    'exit':                    rtn_exit,
    'fclose':                  rtn_fclose,
    'fgets':                   rtn_fgets,
    'fopen':                   rtn_fopen,
    'fprintf':                 rtn_fprintf,
    'fputc':                   rtn_fputc,
    'fputs':                   rtn_fputs,
    'fread':                   rtn_fread,
    'free':                    rtn_free,
    'fwrite':                  rtn_fwrite,
    'gettimeofday':            rtn_gettimeofday,
    'malloc':                  rtn_malloc,
    'memcmp':                  rtn_memcmp,
    'memcpy':                  rtn_memcpy,
    'memmem':                  rtn_memmem,
    'memmove':                 rtn_memmove,
    'memset':                  rtn_memset,
    'open':                    rtn_open,
    'printf':                  rtn_printf,
    'pthread_create':          rtn_pthread_create,
    'pthread_exit':            rtn_pthread_exit,
    'pthread_join':            rtn_pthread_join,
    'pthread_mutex_destroy':   rtn_pthread_mutex_destroy,
    'pthread_mutex_init':      rtn_pthread_mutex_init,
    'pthread_mutex_lock':      rtn_pthread_mutex_lock,
    'pthread_mutex_unlock':    rtn_pthread_mutex_unlock,
    'puts':                    rtn_puts,
    'rand':                    rtn_rand,
    'read':                    rtn_read,
    'sem_destroy':             rtn_sem_destroy,
    'sem_getvalue':            rtn_sem_getvalue,
    'sem_init':                rtn_sem_init,
    'sem_post':                rtn_sem_post,
    'sem_timedwait':           rtn_sem_timedwait,
    'sem_trywait':             rtn_sem_trywait,
    'sem_wait':                rtn_sem_wait,
    'sleep':                   rtn_sleep,
    'sprintf':                 rtn_sprintf,
    'strcasecmp':              rtn_strcasecmp,
    'strchr':                  rtn_strchr,
    'strcmp':                  rtn_strcmp,
    'strcpy':                  rtn_strcpy,
    'strerror':                rtn_strerror,
    'strlen':                  rtn_strlen,
    'strncasecmp':             rtn_strncasecmp,
    'strncmp':                 rtn_strncmp,
    'strncpy':                 rtn_strncpy,
    'strtok_r':                rtn_strtok_r,
    'strtoul':                 rtn_strtoul,

    'write':                   rtn_write,
    'getenv':                  rtn_getenv,
    'fseek':                   rtn_fseek,
    'ftell':                   rtn_ftell,

    '_setjmp':                 rtn__setjmp,
    'longjmp':                 rtn_longjmp,
    'realloc':                 rtn_realloc,
    'setlocale':               rtn_setlocale,
    'strdup':                  rtn_strdup,
    'mempcpy':                 rtn_mempcpy,
    '__mempcpy':               rtn_mempcpy,
    'getchar':                 rtn_getchar,

    'isspace':                 rtn_isspace,
    #'tolower':                 rtn_tolower,
}


SUPORTED_GVARIABLES = {
    '__stack_chk_guard':    0xdead,
    'stderr':               0x0002,
    'stdin':                0x0000,
    'stdout':               0x0001,
}

